//       _________ __                 __
//      /   _____//  |_____________ _/  |______     ____  __ __  ______
//      \_____  \\   __\_  __ \__  \\   __\__  \   / ___\|  |  \/  ___/
//      /        \|  |  |  | \// __ \|  |  / __ \_/ /_/  >  |  /\___ |
//     /_______  /|__|  |__|  (____  /__| (____  /\___  /|____//____  >
//             \/                  \/          \//_____/            \/
//  ______________________                           ______________________
//                        T H E   W A R   B E G I N S
//         Stratagus - A free fantasy real time strategy game engine
//
//      (c) Copyright 1998-2021 by Lutz Sammer, Jimmy Salmon, Nehal Mistry
//		and Andrettin
//
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; only version 2 of the License.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//      02111-1307, USA.

#include "stratagus.h"

#include "player.h"

#include "action/action_upgradeto.h"
#include "actions.h"
#include "age.h"
#include "ai.h"
#include "ai/ai_local.h" //for using AiHelpers
#include "civilization.h"
#include "civilization_group.h"
#include "civilization_history.h"
#include "commands.h" //for faction setting
#include "currency.h"
#include "database/defines.h"
#include "dialogue.h"
#include "diplomacy_state.h"
#include "dynasty.h"
#include "economy/resource_storage_type.h"
#include "editor.h"
#include "faction.h"
#include "faction_history.h"
#include "faction_type.h"
//Wyrmgus start
#include "game.h"
//Wyrmgus end
#include "gender.h"
//Wyrmgus start
#include "grand_strategy.h"
#include "iocompat.h"
//Wyrmgus end
#include "iolib.h"
#include "item/unique_item.h"
#include "language/language.h"
//Wyrmgus start
#include "luacallback.h"
//Wyrmgus end
#include "map/landmass.h"
#include "map/map.h"
#include "map/map_layer.h"
#include "map/minimap.h"
#include "map/site.h"
#include "map/site_game_data.h"
#include "map/tile_flag.h"
#include "network.h"
#include "netconnect.h"
//Wyrmgus start
#include "parameters.h"
//Wyrmgus end
#include "player_color.h"
#include "quest/campaign.h"
#include "quest/objective/quest_objective.h"
#include "quest/objective_type.h"
#include "quest/player_quest_objective.h"
#include "quest/quest.h"
//Wyrmgus start
#include "religion/deity.h"
#include "religion/religion.h"
//Wyrmgus end
#include "script.h"
#include "script/condition/and_condition.h"
#include "script/context.h"
#include "script/effect/effect_list.h"
//Wyrmgus start
#include "settings.h"
#include "sound/sound.h"
#include "sound/unitsound.h"
#include "time/calendar.h"
#include "time/time_of_day.h"
#include "translate.h"
#include "ui/button.h"
#include "ui/button_cmd.h"
#include "ui/ui.h"
#include "unit/unit.h"
//Wyrmgus start
#include "unit/unit_class.h"
#include "unit/unit_find.h"
#include "unit/unit_type.h"
//Wyrmgus end
#include "unit/unit_type_type.h"
//Wyrmgus start
#include "upgrade/upgrade.h"
//Wyrmgus end
#include "upgrade/upgrade_class.h"
#include "upgrade/upgrade_modifier.h"
#include "util/log_util.h"
#include "util/string_util.h"
#include "util/vector_util.h"
#include "util/vector_random_util.h"
#include "vassalage_type.h"
#include "video/font.h"
#include "video/video.h"

/**
**  @class CPlayer player.h
**
**  \#include "player.h"
**
**  This structure contains all information about a player in game.
**
**  The player structure members:
**
**  CPlayer::Player
**
**    This is the unique slot number. It is not possible that two
**    players have the same slot number at the same time. The slot
**    numbers are reused in the future. This means if a player is
**    defeated, a new player can join using this slot. Currently
**    #PlayerMax (16) players are supported. This member is used to
**    access bit fields.
**    Slot #PlayerNumNeutral (15) is reserved for the neutral units
**    like gold-mines or critters.
**
**    @note Should call this member Slot?
**
**  CPlayer::Name
**
**    Name of the player used for displays and network game.
**    It is restricted to 15 characters plus final zero.
**
**  CPlayer::Type
**
**    Type of the player. This field is setup from the level (map).
**    We support currently #PlayerNeutral,
**    #PlayerNobody, #PlayerComputer, #PlayerPerson,
**    #PlayerRescuePassive and #PlayerRescueActive.
**    @see #PlayerTypes.
**
**  CPlayer::RaceName
**
**    Name of the race to which the player belongs, used to select
**    the user interface and the AI.
**    We have 'orc', 'human', 'alliance' or 'mythical'. Should
**    only be used during configuration and not during runtime.
**
**  CPlayer::Race
**
**    Race number of the player. This field is setup from the level
**    map. This number is mapped with #PlayerRaces to the symbolic
**    name CPlayer::RaceName.
**
**  CPlayer::AiName
**
**    AI name for computer. This field is setup
**    from the map. Used to select the AI for the computer
**    player.
**
**  CPlayer::Team
**
**    Team of player. Selected during network game setup. All players
**    of the same team are allied and enemy to all other teams.
**    @note It is planned to show the team on the map.
**
**  CPlayer::Enemy
**
**    A bit field which contains the enemies of this player.
**    If CPlayer::Enemy & (1<<CPlayer::Player) != 0 its an enemy.
**    Setup during startup using the CPlayer::Team, can later be
**    changed with diplomacy. CPlayer::Enemy and CPlayer::Allied
**    are combined, if none bit is set, the player is neutral.
**    @note You can be allied to a player, which sees you as enemy.
**
**  CPlayer::Allied
**
**    A bit field which contains the allies of this player.
**    If CPlayer::Allied & (1<<CPlayer::Player) != 0 its an allied.
**    Setup during startup using the Player:Team, can later be
**    changed with diplomacy. CPlayer::Enemy and CPlayer::Allied
**    are combined, if none bit is set, the player is neutral.
**    @note You can be allied to a player, which sees you as enemy.
**
**  CPlayer::SharedVision
**
**    A bit field which contains shared vision for this player.
**    Shared vision only works when it's activated both ways. Really.
**
**  CPlayer::StartX CPlayer::StartY
**
**    The tile map coordinates of the player start position. 0,0 is
**    the upper left on the map. This members are setup from the
**    map and only important for the game start.
**    Ignored if game starts with level settings. Used to place
**    the initial workers if you play with 1 or 3 workers.
**
**  CPlayer::Resources[::MaxCosts]
**
**    How many resources the player owns. Needed for building
**    units and structures.
**    @see _costs_, TimeCost, GoldCost, WoodCost, OilCost, MaxCosts.
**
**  CPlayer::MaxResources[::MaxCosts]
**
**    How many resources the player can store at the moment.
**
**  CPlayer::Incomes[::MaxCosts]
**
**    Income of the resources, when they are delivered at a store.
**    @see _costs_, TimeCost, GoldCost, WoodCost, OilCost, MaxCosts.
**
**  CPlayer::LastResources[::MaxCosts]
**
**    Keeps track of resources in time (used for calculating
**    CPlayer::Revenue, see below)
**
**  CPlayer::Revenue[::MaxCosts]
**
**    Production of resources per minute (or estimates)
**    Used just as information (statistics) for the player...
**
**  CPlayer::UnitTypesCount[::UnitTypeMax]
**
**    Total count for each different unit type. Used by the AI and
**    for condition checks. The addition of all counts should
**    be CPlayer::TotalNumUnits.
**    @note Should not use the maximum number of unit-types here,
**    only the real number of unit-types used.
**
**  CPlayer::AiEnabled
**
**    If the player is controlled by the computer and this flag is
**    true, than the player is handled by the AI on this local
**    computer.
**
**    @note Currently the AI is calculated parallel on all computers
**    in a network play. It is planned to change this.
**
**  CPlayer::Ai
**
**    AI structure pointer. Please look at #PlayerAi for more
**    information.
**
**  CPlayer::Units
**
**    A table of all (CPlayer::TotalNumUnits) units of the player.
**
**  CPlayer::TotalNumUnits
**
**    Total number of units (incl. buildings) in the CPlayer::Units
**    table.
**
**  CPlayer::Demand
**
**    Total unit demand, used to demand limit.
**    A player can only build up to CPlayer::Food units and not more
**    than CPlayer::FoodUnitLimit units.
**
**    @note that CPlayer::NumFoodUnits > CPlayer::Food, when enough
**    farms are destroyed.
**
**  CPlayer::NumBuildings
**
**    Total number buildings, units that don't need food.
**
**  CPlayer::Food
**
**    Number of food available/produced. Player can't train more
**    CPlayer::NumFoodUnits than this.
**    @note that all limits are always checked.
**
**  CPlayer::FoodUnitLimit
**
**    Number of food units allowed. Player can't train more
**    CPlayer::NumFoodUnits than this.
**    @note that all limits are always checked.
**
**  CPlayer::BuildingLimit
**
**    Number of buildings allowed.  Player can't build more
**    CPlayer::NumBuildings than this.
**    @note that all limits are always checked.
**
**  CPlayer::TotalUnitLimit
**
**    Number of total units allowed. Player can't have more
**    CPlayer::NumFoodUnits+CPlayer::NumBuildings=CPlayer::TotalNumUnits
**    this.
**    @note that all limits are always checked.
**
**  CPlayer::Score
**
**    Total number of points. You can get points for killing units,
**    destroying buildings ...
**
**  CPlayer::TotalUnits
**
**    Total number of units made.
**
**  CPlayer::TotalBuildings
**
**    Total number of buildings made.
**
**  CPlayer::TotalResources[::MaxCosts]
**
**    Total number of resources collected.
**    @see _costs_, TimeCost, GoldCost, WoodCost, OilCost, MaxCosts.
**
**  CPlayer::TotalRazings
**
**    Total number of buildings destroyed.
**
**  CPlayer::TotalKills
**
**    Total number of kills.
**
**  CPlayer::Color
**
**    Color of units of this player on the minimap. Index number
**    into the global palette.
**
**  CPlayer::Allow
**
**    Contains which unit-types and upgrades are allowed for the
**    player. Possible values are:
**    @li  'A' -- allowed,
**    @li  'F' -- forbidden,
**    @li  'R' -- acquired, perhaps other values
**    @li  'Q' -- acquired but forbidden (does it make sense?:))
**    @li  'E' -- enabled, allowed by level but currently forbidden
**    @see CAllow
**
**  CPlayer::UpgradeTimers
**
**    Timer for the upgrades. One timer for all possible upgrades.
**    Initial 0 counted up by the upgrade action, until it reaches
**    the upgrade time.
**    @see _upgrade_timers_
**    @note it is planned to combine research for faster upgrades.
*/

int NumPlayers; //how many player slots used

CPlayer *CPlayer::ThisPlayer = nullptr;
std::vector<CPlayer *> CPlayer::Players; //all players in play

PlayerRace PlayerRaces; //player races

bool NoRescueCheck; //disable rescue check

/**
**  "Translate" (that is, adapt) a proper name from one culture (civilization) to another.
*/
std::string PlayerRace::TranslateName(const std::string &name, const wyrmgus::language *language)
{
	std::string new_name;
	
	if (!language || name.empty()) {
		return new_name;
	}

	// try to translate the entire name, as a particular translation for it may exist
	const auto find_iterator = language->NameTranslations.find(name);
	if (find_iterator != language->NameTranslations.end()) {
		return find_iterator->second[SyncRand(find_iterator->second.size())];
	}
	
	//if adapting the entire name failed, try to match prefixes and suffixes
	if (name.size() > 1) {
		if (name.find(" ") == std::string::npos) {
			for (size_t i = 0; i < name.size(); ++i) {
				std::string name_prefix = name.substr(0, i + 1);
				std::string name_suffix = CapitalizeString(name.substr(i + 1, name.size() - (i + 1)));
			
	//			fprintf(stdout, "Trying to match prefix \"%s\" and suffix \"%s\" for translating name \"%s\" to the \"%s\" language.\n", name_prefix.c_str(), name_suffix.c_str(), name.c_str(), language->Ident.c_str());
			
				const auto prefix_find_iterator = language->NameTranslations.find(name_prefix);
				const auto suffix_find_iterator = language->NameTranslations.find(name_suffix);
				if (prefix_find_iterator != language->NameTranslations.end() && suffix_find_iterator != language->NameTranslations.end()) { // if both a prefix and suffix have been matched
					name_prefix = prefix_find_iterator->second[SyncRand(prefix_find_iterator->second.size())];
					name_suffix = suffix_find_iterator->second[SyncRand(suffix_find_iterator->second.size())];
					name_suffix = DecapitalizeString(name_suffix);
					if (name_prefix.substr(name_prefix.size() - 2, 2) == "gs" && name_suffix.substr(0, 1) == "g") { //if the last two characters of the prefix are "gs", and the first character of the suffix is "g", then remove the final "s" from the prefix (as in "Kniggrtz")
						name_prefix = FindAndReplaceStringEnding(name_prefix, "gs", "g");
					}
					if (name_prefix.substr(name_prefix.size() - 1, 1) == "s" && name_suffix.substr(0, 1) == "s") { //if the prefix ends in "s" and the suffix begins in "s" as well, then remove the final "s" from the prefix (as in "Josefstadt", "Kronstadt" and "Leopoldstadt")
						name_prefix = FindAndReplaceStringEnding(name_prefix, "s", "");
					}

					return name_prefix + name_suffix;
				}
			}
		} else { // if the name contains a space, try to translate each of its elements separately
			size_t previous_pos = 0;
			new_name = name;
			for (size_t i = 0; i < name.size(); ++i) {
				if ((i + 1) == name.size() || name[i + 1] == ' ') {
					std::string name_element = TranslateName(name.substr(previous_pos, i + 1 - previous_pos), language);
				
					if (name_element.empty()) {
						new_name = "";
						break;
					}
				
					new_name = FindAndReplaceString(new_name, name.substr(previous_pos, i + 1 - previous_pos), name_element);
				
					previous_pos = i + 2;
				}
			}
		}
	}
	
	return new_name;
}
//Wyrmgus end

/**
**  Init players.
*/
void InitPlayers()
{
	for (int p = 0; p < PlayerMax; ++p) {
		CPlayer::Players[p]->Index = p;
		if (!CPlayer::Players[p]->Type) {
			CPlayer::Players[p]->Type = PlayerNobody;
		}
	}
}

/**
**  Clean up players.
*/
void CleanPlayers()
{
	CPlayer::SetThisPlayer(nullptr);
	CPlayer::revealed_players.clear();
	for (unsigned int i = 0; i < PlayerMax; ++i) {
		CPlayer::Players[i]->Clear();
	}
	NumPlayers = 0;
	NoRescueCheck = false;
}

/**
**  Save state of players to file.
**
**  @param file  Output file.
**
**  @note FIXME: Not completely saved.
*/
void SavePlayers(CFile &file)
{
	file.printf("\n--------------------------------------------\n");
	file.printf("--- MODULE: players\n\n");

	//  Dump all players
	for (int i = 0; i < NumPlayers; ++i) {
		CPlayer::Players[i]->Save(file);
	}

	file.printf("SetThisPlayer(%d)\n\n", CPlayer::GetThisPlayer()->Index);
}

CPlayer::CPlayer()
{
}

CPlayer::~CPlayer()
{
}

void CPlayer::SetThisPlayer(CPlayer *player)
{
	if (player == CPlayer::GetThisPlayer()) {
		return;
	}

	CPlayer::ThisPlayer = player;
}

CPlayer *CPlayer::GetThisPlayer()
{
	return CPlayer::ThisPlayer;
}

CPlayer *CPlayer::GetPlayer(const int index)
{
	if (index < 0) {
		fprintf(stderr, "Cannot get player for index %d: the index is negative.\n", index);
		return nullptr;
	}

	if (index >= PlayerMax) {
		fprintf(stderr, "Cannot get player for index %d: the maximum value is %d.\n", index, PlayerMax);
		return nullptr;
	}

	return CPlayer::Players[index];
}

const QColor &CPlayer::get_minimap_color() const
{
	return this->get_player_color()->get_colors().at(wyrmgus::defines::get()->get_minimap_color_index());
}

void CPlayer::set_revealed(const bool revealed)
{
	if (revealed == this->is_revealed()) {
		return;
	}

	this->revealed = revealed;

	if (revealed) {
		CPlayer::revealed_players.push_back(this);
	} else {
		wyrmgus::vector::remove(CPlayer::revealed_players, this);
	}
}

void CPlayer::Save(CFile &file) const
{
	const CPlayer &p = *this;
	file.printf("Player(%d,\n", this->Index);
	//Wyrmgus start
	file.printf(" \"race\", \"%s\",", wyrmgus::civilization::get_all()[p.Race]->get_identifier().c_str());
	if (p.Faction != -1) {
		file.printf(" \"faction\", %d,", p.Faction);
	}
	if (p.get_faction_tier() != wyrmgus::faction_tier::none) {
		file.printf(" \"faction-tier\", \"%s\",", wyrmgus::faction_tier_to_string(this->get_faction_tier()).c_str());
	}
	if (p.get_government_type() != wyrmgus::government_type::none) {
		file.printf(" \"government-type\", \"%s\",", wyrmgus::government_type_to_string(this->get_government_type()).c_str());
	}
	if (p.get_dynasty() != nullptr) {
		file.printf(" \"dynasty\", \"%s\",", p.get_dynasty()->get_identifier().c_str());
	}
	if (p.age) {
		file.printf(" \"age\", \"%s\",", p.age->get_identifier().c_str());
	}
	if (p.get_player_color() != nullptr) {
		file.printf(" \"player-color\", \"%s\",", p.get_player_color()->get_identifier().c_str());
	}
	//Wyrmgus end
	file.printf("  \"name\", \"%s\",\n", p.Name.c_str());
	file.printf("  \"type\", ");
	switch (p.Type) {
		case PlayerNeutral:       file.printf("\"neutral\",");         break;
		case PlayerNobody:        file.printf("\"nobody\",");          break;
		case PlayerComputer:      file.printf("\"computer\",");        break;
		case PlayerPerson:        file.printf("\"person\",");          break;
		case PlayerRescuePassive: file.printf("\"rescue-passive\","); break;
		case PlayerRescueActive:  file.printf("\"rescue-active\","); break;
		default:                  file.printf("%d,", p.Type); break;
	}
	//Wyrmgus start
//	file.printf(" \"race\", \"%s\",", PlayerRaces.Name[p.Race].c_str());
	//Wyrmgus end
	file.printf(" \"ai-name\", \"%s\",\n", p.AiName.c_str());
	file.printf("  \"team\", %d,", p.Team);

	file.printf(" \"enemy\", \"");
	for (int j = 0; j < PlayerMax; ++j) {
		file.printf("%c", p.enemies.contains(j) ? 'X' : '_');
	}
	file.printf("\", \"allied\", \"");
	for (int j = 0; j < PlayerMax; ++j) {
		file.printf("%c", p.allies.contains(j) ? 'X' : '_');
	}
	file.printf("\", \"shared-vision\", \"");
	for (int j = 0; j < PlayerMax; ++j) {
		file.printf("%c", p.shared_vision.contains(j) ? 'X' : '_');
	}
	file.printf("\",\n  \"start\", {%d, %d},\n", p.StartPos.x, p.StartPos.y);
	//Wyrmgus start
	file.printf("  \"start-map-layer\", %d,\n", p.StartMapLayer);
	//Wyrmgus end
	if (p.get_overlord() != nullptr) {
		file.printf("  \"overlord\", %d, \"%s\",\n", p.get_overlord()->Index, wyrmgus::vassalage_type_to_string(p.vassalage_type).c_str());
	}

	// Resources
	file.printf("  \"resources\", {");
	for (const auto &[resource, quantity] : p.resources) {
		if (quantity == 0) {
			continue;
		}

		file.printf("\"%s\", %d, ", resource->get_identifier().c_str(), quantity);
	}

	// Stored Resources
	file.printf("},\n  \"stored-resources\", {");
	for (const auto &[resource, quantity] : p.stored_resources) {
		if (quantity == 0) {
			continue;
		}

		file.printf("\"%s\", %d, ", resource->get_identifier().c_str(), quantity);
	}

	// Max Resources
	file.printf("},\n  \"max-resources\", {");
	for (const auto &[resource, quantity] : p.max_resources) {
		file.printf("\"%s\", %d, ", resource->get_identifier().c_str(), quantity);
	}

	// Last Resources
	file.printf("},\n  \"last-resources\", {");
	for (const auto &[resource, quantity] : p.last_resources) {
		if (quantity == 0) {
			continue;
		}

		file.printf("\"%s\", %d, ", resource->get_identifier().c_str(), quantity);
	}

	// Incomes
	file.printf("},\n  \"incomes\", {");
	bool first = true;
	for (const auto &[resource, quantity] : p.incomes) {
		if (first) {
			first = false;
		} else {
			file.printf(" ");
		}
		file.printf("\"%s\", %d,", resource->get_identifier().c_str(), quantity);
	}

	// Revenue
	file.printf("},\n  \"revenue\", {");
	for (const auto &[resource, quantity] : p.revenues) {
		if (quantity == 0) {
			continue;
		}

		file.printf("\"%s\", %d, ", resource->get_identifier().c_str(), quantity);
	}

	file.printf("},\n  \"prices\", {");
	for (const auto &[resource, quantity] : p.prices) {
		if (quantity == 0) {
			continue;
		}

		file.printf("\"%s\", %d, ", resource->get_identifier().c_str(), quantity);
	}

	// UnitTypesCount done by load units.

	file.printf("},\n  \"%s\",\n", p.AiEnabled ? "ai-enabled" : "ai-disabled");

	// Ai done by load ais.
	// Units done by load units.
	// TotalNumUnits done by load units.
	// NumBuildings done by load units.
	
	if (p.is_revealed()) {
		file.printf(" \"revealed\",");
	}
	
	file.printf(" \"supply\", %d,", p.Supply);
	file.printf(" \"trade-cost\", %d,", p.TradeCost);
	file.printf(" \"unit-limit\", %d,", p.UnitLimit);
	file.printf(" \"building-limit\", %d,", p.BuildingLimit);
	file.printf(" \"total-unit-limit\", %d,", p.TotalUnitLimit);

	file.printf("\n  \"score\", %d,", p.Score);
	file.printf("\n  \"total-units\", %d,", p.TotalUnits);
	file.printf("\n  \"total-buildings\", %d,", p.TotalBuildings);
	file.printf("\n  \"total-resources\", {");
	for (const auto &[resource, quantity] : p.resource_totals) {
		if (quantity == 0) {
			continue;
		}

		file.printf("\"%s\", %d, ", resource->get_identifier().c_str(), quantity);
	}
	file.printf("},");
	file.printf("\n  \"total-razings\", %d,", p.TotalRazings);
	file.printf("\n  \"total-kills\", %d,", p.TotalKills);
	//Wyrmgus start
	file.printf("\n  \"unit-type-kills\", {");
	for (const wyrmgus::unit_type *unit_type : wyrmgus::unit_type::get_all()) {
		if (p.UnitTypeKills[unit_type->Slot] != 0) {
			file.printf("\"%s\", %d, ", unit_type->Ident.c_str(), p.UnitTypeKills[unit_type->Slot]);
		}
	}
	file.printf("},");
	//Wyrmgus end
	if (p.LostTownHallTimer != 0) {
		file.printf("\n  \"lost-town-hall-timer\", %d,", p.LostTownHallTimer);
	}
	if (p.HeroCooldownTimer != 0) {
		file.printf("\n  \"hero-cooldown-timer\", %d,", p.HeroCooldownTimer);
	}
	//Wyrmgus end

	file.printf("\n  \"speed-resource-harvest\", {");
	for (const auto &[resource, speed] : p.resource_harvest_speeds) {
		if (speed == CPlayer::base_speed_factor) {
			continue;
		}

		file.printf("\"%s\", %d, ", resource->get_identifier().c_str(), speed);
	}
	file.printf("},");
	file.printf("\n  \"speed-resource-return\", {");
	for (const auto &[resource, speed] : p.resource_return_speeds) {
		if (speed == CPlayer::base_speed_factor) {
			continue;
		}

		file.printf("\"%s\", %d, ", resource->get_identifier().c_str(), speed);
	}
	file.printf("},");
	file.printf("\n  \"speed-build\", %d,", p.SpeedBuild);
	file.printf("\n  \"speed-train\", %d,", p.SpeedTrain);
	file.printf("\n  \"speed-upgrade\", %d,", p.SpeedUpgrade);
	file.printf("\n  \"speed-research\", %d,", p.SpeedResearch);
	
	//Wyrmgus start
	file.printf("\n  \"current-quests\", {");
	for (size_t j = 0; j < p.current_quests.size(); ++j) {
		if (j) {
			file.printf(" ");
		}
		file.printf("\"%s\",", p.current_quests[j]->get_identifier().c_str());
	}
	file.printf("},");
	
	file.printf("\n  \"completed-quests\", {");
	for (size_t j = 0; j < p.completed_quests.size(); ++j) {
		if (j) {
			file.printf(" ");
		}
		file.printf("\"%s\",", p.completed_quests[j]->get_identifier().c_str());
	}
	file.printf("},");
	
	file.printf("\n  \"quest-objectives\", {");
	for (size_t j = 0; j < p.get_quest_objectives().size(); ++j) {
		const auto &objective = p.get_quest_objectives()[j];
		if (j != 0) {
			file.printf(" ");
		}
		file.printf("{");
		file.printf("\"quest\", \"%s\",", objective->get_quest_objective()->get_quest()->get_identifier().c_str());
		file.printf("\"objective-index\", %d,", objective->get_quest_objective()->get_index());
		file.printf("\"counter\", %d,", objective->get_counter());
		file.printf("},");
	}
	file.printf("},");
	
	file.printf("\n  \"modifiers\", {");
	for (size_t j = 0; j < p.Modifiers.size(); ++j) {
		if (j) {
			file.printf(" ");
		}
		file.printf("\"%s\", %d,", p.Modifiers[j].first->get_identifier().c_str(), p.Modifiers[j].second);
	}
	file.printf("},");
	//Wyrmgus end

	file.printf("\n  \"autosell-resources\", {");
	for (size_t j = 0; j < p.AutosellResources.size(); ++j) {
		if (j) {
			file.printf(" ");
		}
		file.printf("\"%s\",", DefaultResourceNames[p.AutosellResources[j]].c_str());
	}
	file.printf("},");
	
	// Allow saved by allow.

	file.printf("\n  \"timers\", {");
	//Wyrmgus start
	first = true;
	//Wyrmgus end
	for (const CUpgrade *upgrade : CUpgrade::get_all()) {
		//Wyrmgus start
//		if (upgrade->ID) {
//			file.printf(" ,");
//		}
//		file.printf("%d", p.UpgradeTimers.Upgrades[upgrade->ID]);
		if (p.UpgradeTimers.Upgrades[upgrade->ID]) {
			if (first) {
				first = false;
			} else {
				file.printf(", ");
			}
			file.printf("\"%s\", %d", upgrade->get_identifier().c_str(), p.UpgradeTimers.Upgrades[upgrade->ID]);
		}
		//Wyrmgus end
	}
	file.printf("}");

	file.printf(")\n\n");

	DebugPrint("FIXME: must save unit-stats?\n");
}

/**
**  Create a new player.
**
**  @param type  Player type (Computer,Human,...).
*/
void CreatePlayer(int type)
{
	if (NumPlayers == PlayerMax) { // already done for bigmaps!
		return;
	}
	CPlayer *player = CPlayer::Players[NumPlayers];
	player->Index = NumPlayers;

	player->Init(type);
}

CPlayer *GetFactionPlayer(const wyrmgus::faction *faction)
{
	if (faction == nullptr) {
		return nullptr;
	}
	
	for (CPlayer *player : CPlayer::Players) {
		if (player->get_faction() == faction) {
			return player;
		}
	}
	
	return nullptr;
}

CPlayer *GetOrAddFactionPlayer(const wyrmgus::faction *faction)
{
	CPlayer *faction_player = GetFactionPlayer(faction);
	if (faction_player != nullptr) {
		return faction_player;
	}
	
	// no player belonging to this faction, so let's make an unused player slot be created for it
	
	for (int i = 0; i < NumPlayers; ++i) {
		CPlayer *player = CPlayer::Players[i];
		if (player->Type == PlayerNobody) {
			player->Type = PlayerComputer;
			player->set_civilization(faction->get_civilization());
			player->SetFaction(faction);
			player->AiEnabled = true;
			player->AiName = faction->DefaultAI;
			player->Team = 1;
			player->set_resource(defines::get()->get_wealth_resource(), 2500); // give the new player enough resources to start up
			player->set_resource(resource::get_all()[WoodCost], 2500);
			player->set_resource(resource::get_all()[StoneCost], 2500);
			return player;
		}
	}
	
	throw std::runtime_error("Cannot add player for faction \"" + faction->get_identifier() + "\": no player slots available.");
}

void CPlayer::Init(/* PlayerTypes */ int type)
{
	std::vector<CUnit *>().swap(this->Units);
	std::vector<CUnit *>().swap(this->FreeWorkers);
	std::vector<CUnit *>().swap(this->LevelUpUnits);

	//  Take first slot for person on this computer,
	//  fill other with computer players.
	if (type == PlayerPerson && !NetPlayers) {
		if (!CPlayer::GetThisPlayer()) {
			CPlayer::SetThisPlayer(this);
		} else {
			type = PlayerComputer;
		}
	}
	if (NetPlayers && NumPlayers == NetLocalPlayerNumber) {
		CPlayer::SetThisPlayer(CPlayer::Players[NetLocalPlayerNumber]);
	}

	if (NumPlayers == PlayerMax) {
		static bool already_warned = false;

		if (!already_warned) {
			DebugPrint("Too many players\n");
			already_warned = true;
		}
		return;
	}

	//  Make simple teams:
	//  All person players are enemies.
	int team;
	switch (type) {
		case PlayerNeutral:
		case PlayerNobody:
		default:
			team = 0;
			this->SetName("Neutral");
			break;
		case PlayerComputer:
			team = 1;
			this->SetName("Computer");
			break;
		case PlayerPerson:
			team = 2 + NumPlayers;
			this->SetName("Person");
			break;
		case PlayerRescuePassive:
		case PlayerRescueActive:
			// FIXME: correct for multiplayer games?
			this->SetName("Computer");
			team = 2 + NumPlayers;
			break;
	}
	DebugPrint("CreatePlayer name %s\n" _C_ this->Name.c_str());

	this->Type = type;
	this->Race = wyrmgus::defines::get()->get_neutral_civilization()->ID;
	this->Faction = -1;
	this->faction_tier = wyrmgus::faction_tier::none;
	this->government_type = wyrmgus::government_type::none;
	this->religion = nullptr;
	this->dynasty = nullptr;
	this->age = nullptr;
	this->overlord = nullptr;
	this->vassalage_type = wyrmgus::vassalage_type::none;
	this->Team = team;
	this->enemies.clear();
	this->allies.clear();
	this->AiName = "ai-passive";

	//  Calculate enemy/allied mask.
	for (int i = 0; i < NumPlayers; ++i) {
		CPlayer *other_player = CPlayer::Players[i];
		switch (type) {
			case PlayerNeutral:
			case PlayerNobody:
			default:
				break;
			case PlayerComputer:
				// Computer allied with computer and enemy of all persons.
				// make computer players be hostile to each other by default
				if (other_player->Type == PlayerComputer || other_player->Type == PlayerPerson || other_player->Type == PlayerRescueActive) {
					this->enemies.insert(i);
					other_player->enemies.insert(NumPlayers);
				}
				break;
			case PlayerPerson:
				// Humans are enemy of all?
				if (other_player->Type == PlayerComputer || other_player->Type == PlayerPerson) {
					this->enemies.insert(i);
					other_player->enemies.insert(NumPlayers);
				} else if (other_player->Type == PlayerRescueActive || other_player->Type == PlayerRescuePassive) {
					this->allies.insert(i);
					other_player->allies.insert(NumPlayers);
				}
				break;
			case PlayerRescuePassive:
				// Rescue passive are allied with persons
				if (other_player->Type == PlayerPerson) {
					this->allies.insert(i);
					other_player->allies.insert(NumPlayers);
				}
				break;
			case PlayerRescueActive:
				// Rescue active are allied with persons and enemies of computer
				if (other_player->Type == PlayerComputer) {
					this->enemies.insert(i);
					other_player->enemies.insert(NumPlayers);
				} else if (other_player->Type == PlayerPerson) {
					this->allies.insert(i);
					other_player->allies.insert(NumPlayers);
				}
				break;
		}
	}

	//initial default incomes.
	for (const resource *resource : resource::get_all()) {
		this->set_income(resource, resource->get_default_income());
	}
	
	this->TradeCost = DefaultTradeCost;

	//initial max resource amounts.
	for (const resource *resource : resource::get_all()) {
		this->max_resources[resource] = resource->DefaultMaxAmount;
	}

	//Wyrmgus start
	this->UnitTypesCount.clear();
	this->UnitTypesUnderConstructionCount.clear();
	this->UnitTypesAiActiveCount.clear();
	this->Heroes.clear();
	this->Deities.clear();
	this->units_by_type.clear();
	this->units_by_class.clear();
	this->AiActiveUnitsByType.clear();
	this->last_created_unit = nullptr;
	//Wyrmgus end

	this->Supply = 0;
	this->Demand = 0;
	this->NumBuildings = 0;
	//Wyrmgus start
	this->NumBuildingsUnderConstruction = 0;
	this->NumTownHalls = 0;
	//Wyrmgus end
	this->Score = 0;
	//Wyrmgus start
	this->LostTownHallTimer = 0;
	this->HeroCooldownTimer = 0;
	//Wyrmgus end

	if (CPlayer::Players[NumPlayers]->Type == PlayerComputer || CPlayer::Players[NumPlayers]->Type == PlayerRescueActive) {
		this->AiEnabled = true;
	} else {
		this->AiEnabled = false;
	}
	this->revealed = false;
	++NumPlayers;
}

void CPlayer::apply_history(const CDate &start_date)
{
	const wyrmgus::civilization *civilization = this->get_civilization();

	this->apply_civilization_history(civilization);

	const wyrmgus::faction *faction = this->get_faction();
	const wyrmgus::faction_history *faction_history = faction->get_history();

	this->set_faction_tier(faction_history->get_tier());
	this->set_government_type(faction_history->get_government_type());
	this->set_dynasty(faction_history->get_dynasty());

	for (const wyrmgus::upgrade_class *upgrade_class : faction_history->get_acquired_upgrade_classes()) {
		const CUpgrade *upgrade = faction->get_class_upgrade(upgrade_class);

		if (upgrade == nullptr) {
			continue;
		}

		if (UpgradeIdAllowed(*this, upgrade->ID) != 'R') {
			UpgradeAcquire(*this, upgrade);
		}
	}

	for (const CUpgrade *upgrade : faction_history->get_acquired_upgrades()) {
		if (UpgradeIdAllowed(*this, upgrade->ID) != 'R') {
			UpgradeAcquire(*this, upgrade);
		}
	}

	for (const auto &kv_pair : faction->HistoricalDiplomacyStates) { //set the appropriate historical diplomacy states to other factions
		if (kv_pair.first.first.Year == 0 || start_date.ContainsDate(kv_pair.first.first)) {
			CPlayer *diplomacy_state_player = GetFactionPlayer(kv_pair.first.second);
			if (diplomacy_state_player) {
				CommandDiplomacy(this->Index, kv_pair.second, diplomacy_state_player->Index);
				CommandDiplomacy(diplomacy_state_player->Index, kv_pair.second, this->Index);
				if (kv_pair.second == wyrmgus::diplomacy_state::allied) {
					CommandSharedVision(this->Index, true, diplomacy_state_player->Index);
					CommandSharedVision(diplomacy_state_player->Index, true, this->Index);
				}
			}
		}
	}

	for (const auto &kv_pair : faction_history->get_diplomacy_states()) {
		const wyrmgus::faction *other_faction = kv_pair.first;
		const wyrmgus::diplomacy_state state = kv_pair.second;

		CPlayer *diplomacy_state_player = GetFactionPlayer(other_faction);
		if (diplomacy_state_player != nullptr) {
			switch (state) {
				case wyrmgus::diplomacy_state::overlord:
				case wyrmgus::diplomacy_state::personal_union_overlord:
				case wyrmgus::diplomacy_state::vassal:
				case wyrmgus::diplomacy_state::personal_union_vassal:
					CommandDiplomacy(this->Index, state, diplomacy_state_player->Index);
					break;
				case wyrmgus::diplomacy_state::allied:
					CommandSharedVision(this->Index, true, diplomacy_state_player->Index);
					CommandSharedVision(diplomacy_state_player->Index, true, this->Index);
					//fallthrough
				default:
					CommandDiplomacy(this->Index, state, diplomacy_state_player->Index);
					CommandDiplomacy(diplomacy_state_player->Index, state, this->Index);
					break;
			}
		}
	}

	for (const auto &kv_pair : faction->HistoricalResources) { //set the appropriate historical resource quantities
		if (kv_pair.first.first.Year == 0 || start_date.ContainsDate(kv_pair.first.first)) {
			this->set_resource(wyrmgus::resource::get_all()[kv_pair.first.second], kv_pair.second, resource_storage_type::overall);
		}
	}

	for (const auto &kv_pair : faction_history->get_resources()) {
		const wyrmgus::resource *resource = kv_pair.first;
		const int quantity = kv_pair.second;
		this->set_resource(resource, quantity, resource_storage_type::overall);
	}
}

void CPlayer::apply_civilization_history(const wyrmgus::civilization_base *civilization)
{
	if (civilization->get_group() != nullptr) {
		this->apply_civilization_history(civilization->get_group());
	}

	const wyrmgus::civilization_history *civilization_history = civilization->get_history();

	const wyrmgus::faction *faction = this->get_faction();

	for (const wyrmgus::upgrade_class *upgrade_class : civilization_history->get_acquired_upgrade_classes()) {
		const CUpgrade *upgrade = faction->get_class_upgrade(upgrade_class);

		if (upgrade == nullptr) {
			continue;
		}

		if (UpgradeIdAllowed(*this, upgrade->ID) != 'R') {
			UpgradeAcquire(*this, upgrade);
		}
	}

	for (const CUpgrade *upgrade : civilization_history->get_acquired_upgrades()) {
		if (UpgradeIdAllowed(*this, upgrade->ID) != 'R') {
			UpgradeAcquire(*this, upgrade);
		}
	}
}

bool CPlayer::is_neutral_player() const
{
	return this->get_index() == PlayerNumNeutral;
}

void CPlayer::SetName(const std::string &name)
{
	this->Name = name;
}

const wyrmgus::civilization *CPlayer::get_civilization() const
{
	if (this->Race != -1) {
		return wyrmgus::civilization::get_all()[this->Race];
	}

	return nullptr;
}

//Wyrmgus start
void CPlayer::set_civilization(const wyrmgus::civilization *civilization)
{
	if (this->get_civilization() != nullptr && (GameRunning || GameEstablishing)) {
		const wyrmgus::civilization *old_civilization = this->get_civilization();
		if (old_civilization->get_upgrade() != nullptr && this->Allow.Upgrades[old_civilization->get_upgrade()->ID] == 'R') {
			UpgradeLost(*this, old_civilization->get_upgrade()->ID);
		}
	}

	if (GameRunning) {
		this->SetFaction(nullptr);
	} else {
		this->Faction = -1;
	}

	this->Race = civilization->ID;

	if (this->get_civilization() != nullptr) {
		//if the civilization of the person player changed, update the UI
		if ((CPlayer::GetThisPlayer() && CPlayer::GetThisPlayer()->Index == this->Index) || (!CPlayer::GetThisPlayer() && this->Index == 0)) {
			//load proper UI
			std::array<char, 256> buf{};
			snprintf(buf.data(), sizeof(buf), "if (LoadCivilizationUI ~= nil) then LoadCivilizationUI(\"%s\") end;", this->get_civilization()->get_identifier().c_str());
			CclCommand(buf.data());

			UI.Load();
		}

		const wyrmgus::civilization *new_civilization = this->get_civilization();
		CUpgrade *civilization_upgrade = new_civilization->get_upgrade();
		if (civilization_upgrade != nullptr && this->Allow.Upgrades[civilization_upgrade->ID] != 'R') {
			UpgradeAcquire(*this, civilization_upgrade);
		}
	}
}

wyrmgus::faction *CPlayer::get_faction() const
{
	if (this->Faction != -1) {
		return wyrmgus::faction::get_all()[this->Faction];
	}

	return nullptr;
}

void CPlayer::SetFaction(const wyrmgus::faction *faction)
{
	if (faction == this->get_faction()) {
		return;
	}

	const wyrmgus::faction *old_faction = this->get_faction();

	if (faction != nullptr && faction->get_civilization() != this->get_civilization()) {
		this->set_civilization(faction->get_civilization());
	}

	if (this->get_faction() != nullptr) {
		if (!this->get_faction()->FactionUpgrade.empty() && this->Allow.Upgrades[CUpgrade::get(this->get_faction()->FactionUpgrade)->ID] == 'R') {
			UpgradeLost(*this, CUpgrade::get(this->get_faction()->FactionUpgrade)->ID);
		}

		const CUpgrade *faction_type_upgrade = wyrmgus::defines::get()->get_faction_type_upgrade(this->get_faction()->get_type());
		if (faction_type_upgrade != nullptr && this->Allow.Upgrades[faction_type_upgrade->ID] == 'R') {
			UpgradeLost(*this, faction_type_upgrade->ID);
		}
	}

	const int faction_id = faction ? faction->ID : -1;
	
	if (old_faction != nullptr && faction != nullptr) {
		for (const wyrmgus::upgrade_class *upgrade_class : wyrmgus::upgrade_class::get_all()) {
			const CUpgrade *old_faction_class_upgrade = old_faction->get_class_upgrade(upgrade_class);
			const CUpgrade *new_faction_class_upgrade = faction->get_class_upgrade(upgrade_class);
			if (old_faction_class_upgrade != new_faction_class_upgrade) { //if the upgrade for a certain class is different for the new faction than the old faction (and it has been acquired), remove the modifiers of the old upgrade and apply the modifiers of the new
				if (old_faction_class_upgrade != nullptr && this->Allow.Upgrades[old_faction_class_upgrade->ID] == 'R') {
					UpgradeLost(*this, old_faction_class_upgrade->ID);

					if (new_faction_class_upgrade != nullptr) {
						UpgradeAcquire(*this, new_faction_class_upgrade);
					}
				}
			}
		}
	}
	
	bool personal_names_changed = true;
	if (this->get_faction() != nullptr && faction != nullptr) {
		//setting to a faction of the same civilization (if the civilization were different, then the set_civilization() call above would have led to the current faction becoming null)
		personal_names_changed = false;
	}
	
	this->Faction = faction_id;

	if (this->Index == CPlayer::GetThisPlayer()->Index) {
		UI.Load();
	}
	
	if (this->get_faction() == nullptr) {
		return;
	}
	
	if (!IsNetworkGame()) { //only set the faction's name as the player's name if this is a single player game
		this->SetName(this->get_faction()->get_name());
	}
	if (this->get_faction() != nullptr) {
		const wyrmgus::player_color *player_color = nullptr;

		this->set_faction_tier(faction->get_default_tier());
		this->set_government_type(faction->get_default_government_type());

		const wyrmgus::player_color *faction_color = faction->get_color();
		if (faction_color != nullptr) {
			if (this->get_player_color_usage_count(faction_color) == 0) {
				player_color = faction_color;
			}
		}

		if (player_color == nullptr) {
			//if all of the faction's colors are used, get one of the least used player colors
			//out of those colors, give priority to the one closest (in RGB values) to the faction's color
			int best_usage_count = -1;
			int best_rgb_difference = -1;
			std::vector<const wyrmgus::player_color *> available_colors;
			for (const wyrmgus::player_color *pc : wyrmgus::player_color::get_all()) {
				if (pc == wyrmgus::defines::get()->get_neutral_player_color()) {
					continue;
				}

				if (pc->is_hidden()) {
					continue;
				}

				const int usage_count = this->get_player_color_usage_count(pc);
				if (best_usage_count != -1 && usage_count > best_usage_count) {
					continue;
				}

				if (usage_count < best_usage_count) {
					available_colors.clear();
					best_rgb_difference = -1;
				}

				int rgb_difference = -1;

				if (faction_color != nullptr) {
					for (size_t i = 0; i < faction_color->get_colors().size(); ++i) {
						const QColor &faction_color_shade = faction_color->get_colors()[i];
						const QColor &pc_color_shade = pc->get_colors()[i];
						rgb_difference += std::abs(faction_color_shade.red() - pc_color_shade.red());
						rgb_difference += std::abs(faction_color_shade.green() - pc_color_shade.green());
						rgb_difference += std::abs(faction_color_shade.blue() - pc_color_shade.blue());
					}
				}

				if (best_rgb_difference != -1 && rgb_difference > best_rgb_difference) {
					continue;
				}

				if (rgb_difference < best_rgb_difference) {
					available_colors.clear();
				}

				available_colors.push_back(pc);
				best_usage_count = usage_count;
				best_rgb_difference = rgb_difference;
			}

			if (!available_colors.empty()) {
				player_color = available_colors[SyncRand(available_colors.size())];
			}
		}
		
		if (player_color == nullptr) {
			throw std::runtime_error("No player color chosen for player \"" + this->Name + "\" (" + std::to_string(this->Index) + ").");
		}

		this->player_color = player_color;

		if (Editor.Running == EditorNotRunning) {
			//update the territory on the minimap for the new color
			this->update_minimap_territory();
		}

		if (!this->get_faction()->FactionUpgrade.empty()) {
			CUpgrade *faction_upgrade = CUpgrade::try_get(this->get_faction()->FactionUpgrade);
			if (faction_upgrade && this->Allow.Upgrades[faction_upgrade->ID] != 'R') {
				if (GameEstablishing) {
					AllowUpgradeId(*this, faction_upgrade->ID, 'R');
				} else {
					UpgradeAcquire(*this, faction_upgrade);
				}
			}
		}
		
		const CUpgrade *faction_type_upgrade = wyrmgus::defines::get()->get_faction_type_upgrade(this->get_faction()->get_type());
		if (faction_type_upgrade != nullptr && this->Allow.Upgrades[faction_type_upgrade->ID] != 'R') {
			if (GameEstablishing) {
				AllowUpgradeId(*this, faction_type_upgrade->ID, 'R');
			} else {
				UpgradeAcquire(*this, faction_type_upgrade);
			}
		}
	} else {
		fprintf(stderr, "Invalid faction \"%s\" tried to be set for player %d of civilization \"%s\".\n", faction->get_name().c_str(), this->Index, this->get_civilization()->get_identifier().c_str());
	}
	
	for (int i = 0; i < this->GetUnitCount(); ++i) {
		CUnit &unit = this->GetUnit(i);

		if (unit.Type->BoolFlag[ORGANIC_INDEX].value) {
			if (personal_names_changed && unit.get_character() == nullptr && unit.Type->get_civilization() != nullptr && unit.Type->get_civilization()->get_species() == faction->get_civilization()->get_species() && unit.Type == faction->get_class_unit_type(unit.Type->get_unit_class())) {
				unit.UpdatePersonalName();
			}
		} else {
			if (unit.get_unique() == nullptr) {
				const unit_class *unit_class = unit.Type->get_unit_class();
				if (unit_class != nullptr && (old_faction == nullptr || faction == nullptr || old_faction->get_unit_class_name_generator(unit_class) != faction->get_unit_class_name_generator(unit_class))) {
					unit.UpdatePersonalName();
				}
			}
		}

		unit.UpdateSoldUnits();
		unit.UpdateButtonIcons();
	}
}

/**
**  Change player faction to a randomly chosen one.
**
**  @param faction    New faction.
*/
void CPlayer::SetRandomFaction()
{
	// set random one from the civilization's factions
	std::vector<wyrmgus::faction *> local_factions;
	
	for (wyrmgus::faction *faction : wyrmgus::faction::get_all()) {
		if (faction->get_civilization() != this->get_civilization()) {
			continue;
		}

		if (!faction->Playable) {
			continue;
		}

		if (!this->can_found_faction(faction)) {
			continue;
		}

		const wyrmgus::faction_type faction_type = faction->get_type();
		const bool has_writing = this->has_upgrade_class(wyrmgus::upgrade_class::get("writing"));
		if (
			!(faction_type == wyrmgus::faction_type::tribe && !has_writing)
			&& !(faction_type == wyrmgus::faction_type::polity && has_writing)
		) {
			continue;
		}

		local_factions.push_back(faction);
	}
	
	if (local_factions.size() > 0) {
		wyrmgus::faction *chosen_faction = local_factions[SyncRand(local_factions.size())];
		this->SetFaction(chosen_faction);
	} else {
		this->SetFaction(nullptr);
	}
}

void CPlayer::set_dynasty(const wyrmgus::dynasty *dynasty)
{
	if (dynasty == this->get_dynasty()) {
		return;
	}

	const wyrmgus::dynasty *old_dynasty = this->dynasty;
	
	if (old_dynasty != nullptr) {
		if (old_dynasty->get_upgrade() != nullptr && this->Allow.Upgrades[old_dynasty->get_upgrade()->ID] == 'R') {
			UpgradeLost(*this, old_dynasty->get_upgrade()->ID);
		}
	}

	this->dynasty = dynasty;

	if (dynasty == nullptr) {
		return;
	}
	
	if (dynasty->get_upgrade() != nullptr) {
		if (this->Allow.Upgrades[dynasty->get_upgrade()->ID] != 'R') {
			if (GameEstablishing) {
				AllowUpgradeId(*this, dynasty->get_upgrade()->ID, 'R');
			} else {
				UpgradeAcquire(*this, dynasty->get_upgrade());
			}
		}
	}

	for (int i = 0; i < this->GetUnitCount(); ++i) {
		CUnit &unit = this->GetUnit(i);
		unit.UpdateSoldUnits(); //in case conditions changed (e.g. some heroes may require a certain dynasty)
	}
}

const std::string &CPlayer::get_interface() const
{
	const wyrmgus::civilization *civilization = this->get_civilization();
	if (civilization != nullptr) {
		return civilization->get_interface();
	}

	return wyrmgus::string::empty_str;
}

/**
**	@brief	Check which age fits the player's current situation best, and set it as the player's age
*/
void CPlayer::check_age()
{
	//pick an age which fits the player, giving priority to the first ones (ages are already sorted by priority)
	
	for (wyrmgus::age *potential_age : wyrmgus::age::get_all()) {
		if (!check_conditions(potential_age, this)) {
			continue;
		}
		
		this->set_age(potential_age);
		return;
	}
	
	this->set_age(nullptr);
}

/**
**	@brief	Set the player's age
**
**	@param	age	The age to be set for the player
*/
void CPlayer::set_age(const wyrmgus::age *age)
{
	if (this->age == age) {
		return;
	}
	
	this->age = age;
	
	if (this == CPlayer::GetThisPlayer()) {
		if (this->age != nullptr) {
			if (GameCycle > 0 && !SaveGameLoading) {
				this->Notify(_("The %s has dawned upon us."), _(this->age->get_name().c_str()));
			}
		}
	}
	
	wyrmgus::age::check_current_age();
}

/**
**	@brief	Get the player's currency
**
**	@return	The player's currency
*/
CCurrency *CPlayer::GetCurrency() const
{
	if (this->get_faction() != nullptr) {
		return this->get_faction()->GetCurrency();
	}
	
	if (this->get_civilization() != nullptr) {
		return this->get_civilization()->GetCurrency();
	}
	
	return nullptr;
}

void CPlayer::ShareUpgradeProgress(CPlayer &player, CUnit &unit)
{
	std::vector<const CUpgrade *> upgrade_list = this->GetResearchableUpgrades();
	std::vector<const CUpgrade *> potential_upgrades;

	for (size_t i = 0; i < upgrade_list.size(); ++i) {
		if (this->Allow.Upgrades[upgrade_list[i]->ID] != 'R') {
			continue;
		}
		
		if (upgrade_list[i]->get_upgrade_class() == nullptr) {
			continue;
		}

		if (player.Faction == -1) {
			continue;
		}
		
		const CUpgrade *upgrade = player.get_faction()->get_class_upgrade(upgrade_list[i]->get_upgrade_class());
		if (upgrade == nullptr) {
			continue;
		}
		
		if (player.Allow.Upgrades[upgrade->ID] != 'A' || !check_conditions(upgrade, &player)) {
			continue;
		}
	
		if (player.UpgradeRemovesExistingUpgrade(upgrade, player.AiEnabled)) {
			continue;
		}
		
		potential_upgrades.push_back(upgrade);
	}
	
	if (potential_upgrades.size() > 0) {
		const CUpgrade *chosen_upgrade = potential_upgrades[SyncRand(potential_upgrades.size())];
		
		if (!chosen_upgrade->get_name().empty()) {
			player.Notify(NotifyGreen, unit.tilePos, unit.MapLayer->ID, _("%s acquired through contact with %s"), chosen_upgrade->get_name().c_str(), this->Name.c_str());
		}
		if (&player == CPlayer::GetThisPlayer() && player.get_civilization() != nullptr) {
			const wyrmgus::sound *sound = player.get_civilization()->get_research_complete_sound();

			if (sound != nullptr) {
				PlayGameSound(sound, MaxSampleVolume);
			}
		}
		if (player.AiEnabled) {
			AiResearchComplete(unit, chosen_upgrade);
		}
		UpgradeAcquire(player, chosen_upgrade);
	}
}

int CPlayer::get_player_color_usage_count(const wyrmgus::player_color *player_color) const
{
	int count = 0;

	for (int i = 0; i < PlayerMax; ++i) {
		if (this->Index != i && CPlayer::Players[i]->Faction != -1 && CPlayer::Players[i]->Type != PlayerNobody && CPlayer::Players[i]->get_player_color() == player_color) {
			count++;
		}		
	}

	return count;
}

void CPlayer::update_minimap_territory()
{
	for (const auto &kv_pair : this->get_units_by_type()) {
		const wyrmgus::unit_type *unit_type = kv_pair.first;
		if (!unit_type->BoolFlag[TOWNHALL_INDEX].value) {
			continue;
		}

		for (const CUnit *town_hall : kv_pair.second) {
			town_hall->settlement->get_game_data()->update_minimap_territory();
		}
	}

	//also update the minimap territory of vassals, as they get strokes of the overlord's colors
	for (CPlayer *vassal : this->get_vassals()) {
		vassal->update_minimap_territory();
	}
}

unit_type *CPlayer::get_class_unit_type(const unit_class *unit_class) const
{
	const faction *faction = this->get_faction();
	if (faction == nullptr) {
		return nullptr;
	}

	return faction->get_class_unit_type(unit_class);
}

bool CPlayer::is_class_unit_type(const unit_type *unit_type) const
{
	const faction *faction = this->get_faction();
	if (faction == nullptr) {
		return false;
	}

	return faction->is_class_unit_type(unit_type);
}

CUpgrade *CPlayer::get_class_upgrade(const wyrmgus::upgrade_class *upgrade_class) const
{
	const wyrmgus::faction *faction = this->get_faction();
	if (faction == nullptr) {
		return nullptr;
	}

	return faction->get_class_upgrade(upgrade_class);
}

bool CPlayer::has_upgrade_class(const wyrmgus::upgrade_class *upgrade_class) const
{
	if (this->get_civilization() == nullptr || upgrade_class == nullptr) {
		return false;
	}
	
	const CUpgrade *upgrade = nullptr;
	
	if (this->get_faction() != nullptr) {
		upgrade = this->get_faction()->get_class_upgrade(upgrade_class);
	} else {
		upgrade = this->get_civilization()->get_class_upgrade(upgrade_class);
	}
	
	if (upgrade != nullptr && this->Allow.Upgrades[upgrade->ID] == 'R') {
		return true;
	}

	return false;
}

std::vector<CUnit *> CPlayer::get_town_hall_units() const
{
	std::vector<CUnit *> town_hall_units;

	for (const auto &kv_pair : this->get_units_by_type()) {
		const wyrmgus::unit_type *unit_type = kv_pair.first;
		if (!unit_type->BoolFlag[TOWNHALL_INDEX].value) {
			continue;
		}

		FindPlayerUnitsByType(*this, *unit_type, town_hall_units, true);
	}

	return town_hall_units;
}

std::vector<const wyrmgus::site *> CPlayer::get_settlements() const
{
	std::vector<const wyrmgus::site *> settlements;

	const std::vector<CUnit *> town_hall_units = this->get_town_hall_units();
	for (const CUnit *town_hall_unit : town_hall_units) {
		settlements.push_back(town_hall_unit->settlement);
	}

	return settlements;
}

bool CPlayer::has_settlement(const wyrmgus::site *settlement) const
{
	if (settlement == nullptr) {
		return false;
	}

	const wyrmgus::site_game_data *settlement_game_data = settlement->get_game_data();

	if (settlement_game_data->get_site_unit() && settlement_game_data->get_site_unit()->Player == this) {
		return true;
	}

	return false;
}

bool CPlayer::has_coastal_settlement() const
{
	const std::vector<const wyrmgus::site *> settlements = this->get_settlements();

	for (const wyrmgus::site *settlement : settlements) {
		if (settlement->get_game_data()->is_coastal()) {
			return true;
		}
	}

	return false;
}

bool CPlayer::HasSettlementNearWaterZone(const landmass *water_zone) const
{
	const std::vector<CUnit *> settlement_unit_table = this->get_town_hall_units();

	for (const CUnit *settlement_unit : settlement_unit_table) {
		if (!settlement_unit->IsAliveOnMap()) {
			continue;
		}
		
		const landmass *settlement_landmass = CMap::Map.get_tile_landmass(settlement_unit->tilePos, settlement_unit->MapLayer->ID);
		if (settlement_landmass == nullptr || !settlement_landmass->borders_landmass(water_zone)) {
			//settlement's landmass doesn't even border the water zone, continue
			continue;
		}
		
		Vec2i pos(0, 0);
		if (FindTerrainType(tile_flag::none, nullptr, 8, *this, settlement_unit->tilePos, &pos, settlement_unit->MapLayer->ID, water_zone)) {
			return true;
		}
	}

	return false;
}

bool CPlayer::has_settlement_with_resource_source(const wyrmgus::resource *resource) const
{
	const std::vector<const wyrmgus::site *> settlements = this->get_settlements();

	for (const wyrmgus::site *settlement : settlements) {
		if (settlement->get_game_data()->has_resource_source(resource)) {
			return true;
		}
	}

	return false;
}

const wyrmgus::site *CPlayer::GetNearestSettlement(const Vec2i &pos, int z, const Vec2i &size) const
{
	CUnit *best_hall = nullptr;
	int best_distance = -1;
	
	for (CUnit *settlement_unit : CMap::Map.get_settlement_units()) {
		if (!settlement_unit || !settlement_unit->IsAliveOnMap() || !settlement_unit->Type->BoolFlag[TOWNHALL_INDEX].value || z != settlement_unit->MapLayer->ID) {
			continue;
		}
		if (!this->has_neutral_faction_type() && this != settlement_unit->Player) {
			continue;
		}
		int distance = MapDistance(size, pos, z, settlement_unit->Type->get_tile_size(), settlement_unit->tilePos, settlement_unit->MapLayer->ID);
		if (!best_hall || distance < best_distance) {
			best_hall = settlement_unit;
			best_distance = distance;
		}
	}
	
	if (best_hall) {
		return best_hall->settlement;
	} else {
		return nullptr;
	}
}

void CPlayer::update_building_settlement_assignment(const wyrmgus::site *old_settlement, const int z) const
{
	for (int i = 0; i < this->GetUnitCount(); ++i) {
		CUnit *unit = &this->GetUnit(i);

		if (unit == nullptr || !unit->IsAliveOnMap()) {
			continue;
		}

		if (!unit->Type->BoolFlag[BUILDING_INDEX].value || unit->Type->BoolFlag[TOWNHALL_INDEX].value || unit->Type == settlement_site_unit_type || unit->MapLayer->ID != z) {
			continue;
		}

		if (old_settlement != nullptr && unit->settlement != old_settlement) {
			continue;
		}

		unit->UpdateSettlement();
	}
}

bool CPlayer::HasUnitBuilder(const wyrmgus::unit_type *type, const wyrmgus::site *settlement) const
{
	const std::vector<const wyrmgus::unit_type *> *builders = nullptr;
	const std::vector<const wyrmgus::unit_class *> *builder_classes = nullptr;

	if (type->BoolFlag[BUILDING_INDEX].value) {
		builders = &AiHelpers.get_builders(type);
		builder_classes = &AiHelpers.get_builder_classes(type->get_unit_class());
	} else {
		builders = &AiHelpers.get_trainers(type);
		builder_classes = &AiHelpers.get_trainer_classes(type->get_unit_class());
	}

	for (const wyrmgus::unit_type *builder : *builders) {
		if (this->GetUnitTypeCount(builder) > 0) {
			return true;
		}
	}

	if (this->get_faction() != nullptr) {
		for (const wyrmgus::unit_class *builder_class : *builder_classes) {
			const wyrmgus::unit_type *builder = this->get_faction()->get_class_unit_type(builder_class);

			if (builder == nullptr) {
				continue;
			}

			if (this->GetUnitTypeCount(builder) > 0) {
				return true;
			}
		}
	}

	const std::vector<const wyrmgus::unit_type *> &unit_type_upgradees = AiHelpers.get_unit_type_upgradees(type);
	const std::vector<const wyrmgus::unit_class *> &unit_class_upgradees = AiHelpers.get_unit_class_upgradees(type->get_unit_class());

	for (const wyrmgus::unit_type *unit_type_upgradee : unit_type_upgradees) {
		if (this->GetUnitTypeCount(unit_type_upgradee) > 0) {
			if (!settlement) {
				return true;
			} else {
				for (int i = 0; i < this->GetUnitCount(); ++i) {
					CUnit &unit = this->GetUnit(i);
					if (unit.Type == unit_type_upgradee && unit.settlement == settlement) {
						return true;
					}
				}
			}
		}
	}

	if (this->get_faction() != nullptr) {
		for (const wyrmgus::unit_class *unit_class_upgradee : unit_class_upgradees) {
			const wyrmgus::unit_type *unit_type_upgradee = this->get_faction()->get_class_unit_type(unit_class_upgradee);

			if (unit_type_upgradee == nullptr) {
				continue;
			}

			if (this->GetUnitTypeCount(unit_type_upgradee) > 0) {
				if (!settlement) {
					return true;
				} else {
					for (int i = 0; i < this->GetUnitCount(); ++i) {
						CUnit &unit = this->GetUnit(i);
						if (unit.Type->get_unit_class() == unit_class_upgradee && unit.settlement == settlement) {
							return true;
						}
					}
				}
			}
		}
	}

	return false;
}

bool CPlayer::HasUpgradeResearcher(const CUpgrade *upgrade) const
{
	for (const wyrmgus::unit_type *researcher_type : AiHelpers.get_researchers(upgrade)) {
		if (this->GetUnitTypeCount(researcher_type) > 0 || HasUnitBuilder(researcher_type)) {
			return true;
		}
	}

	for (const wyrmgus::unit_class *researcher_class : AiHelpers.get_researcher_classes(upgrade->get_upgrade_class())) {
		const wyrmgus::unit_type *researcher_type = this->get_class_unit_type(researcher_class);
		if (researcher_type != nullptr && (this->GetUnitTypeCount(researcher_type) > 0 || this->HasUnitBuilder(researcher_type))) {
			return true;
		}
	}

	return false;
}

template <bool preconditions_only>
bool CPlayer::can_found_faction(const wyrmgus::faction *faction) const
{
	if (CurrentQuest != nullptr) {
		return false;
	}
	
	const CUpgrade *faction_type_upgrade = wyrmgus::defines::get()->get_faction_type_upgrade(faction->get_type());
	if (faction_type_upgrade != nullptr && !wyrmgus::check_conditions<preconditions_only>(faction_type_upgrade, this, false)) {
		return false;
	}

	if (!faction->FactionUpgrade.empty()) {
		const CUpgrade *faction_upgrade = CUpgrade::get(faction->FactionUpgrade);
		if (!wyrmgus::check_conditions<preconditions_only>(faction_upgrade, this, false)) {
			return false;
		}
	}

	for (int i = 0; i < PlayerMax; ++i) {
		if (this->Index != i && CPlayer::Players[i]->Type != PlayerNobody && CPlayer::Players[i]->Race == faction->get_civilization()->ID && CPlayer::Players[i]->Faction == faction->ID) {
			// faction is already in use
			return false;
		}
	}

	if (faction->get_preconditions() != nullptr && !faction->get_preconditions()->check(this)) {
		return false;
	}

	if constexpr (!preconditions_only) {
		//check if the required core settlements are owned by the player
		if (wyrmgus::game::get()->get_current_campaign() != nullptr) { //only check for settlements in the Scenario mode
			for (const wyrmgus::site *core_settlement : faction->get_core_settlements()) {
				const wyrmgus::site_game_data *settlement_game_data = core_settlement->get_game_data();

				if (settlement_game_data->get_site_unit() == nullptr || settlement_game_data->get_site_unit()->Player != this || settlement_game_data->get_site_unit()->CurrentAction() == UnitAction::Built) {
					return false;
				}
			}
		}

		if (faction->get_conditions() != nullptr && !faction->get_conditions()->check(this)) {
			return false;
		}
	}
	
	return true;
}

template bool CPlayer::can_found_faction<false>(const wyrmgus::faction *faction) const;
template bool CPlayer::can_found_faction<true>(const wyrmgus::faction *faction) const;

template <bool preconditions_only>
bool CPlayer::can_choose_dynasty(const wyrmgus::dynasty *dynasty) const
{
	if (CurrentQuest != nullptr) {
		return false;
	}
	
	if (dynasty->get_upgrade() == nullptr) {
		return false;
	}

	if (!wyrmgus::check_conditions<preconditions_only>(dynasty->get_upgrade(), this, false)) {
		return false;
	}

	return wyrmgus::check_conditions<preconditions_only>(dynasty, this, false);
}

template bool CPlayer::can_choose_dynasty<false>(const wyrmgus::dynasty *dynasty) const;
template bool CPlayer::can_choose_dynasty<true>(const wyrmgus::dynasty *dynasty) const;


bool CPlayer::is_character_available_for_recruitment(const wyrmgus::character *character, bool ignore_neutral) const
{
	if (character->is_deity()) {
		return false;
	}
	
	if (!character->get_civilization() || character->get_civilization() != this->get_civilization()) {
		return false;
	}
	
	if (character->get_conditions() != nullptr && !character->get_conditions()->check(this)) {
		return false;
	}
	
	//the conditions for the character's unit type must be fulfilled
	if (!check_conditions(character->get_unit_type(), this, true)) {
		return false;
	}
	
	if (character->Conditions) {
		CclCommand("trigger_player = " + std::to_string(this->Index) + ";");
		character->Conditions->pushPreamble();
		character->Conditions->run(1);
		if (character->Conditions->popBoolean() == false) {
			return false;
		}
	}
	
	if (!character->CanAppear(ignore_neutral)) {
		return false;
	}
	
	return true;
}

std::vector<wyrmgus::character *> CPlayer::get_recruitable_heroes_from_list(const std::vector<wyrmgus::character *> &heroes)
{
	std::vector<wyrmgus::character *> recruitable_heroes;

	for (wyrmgus::character *hero : heroes) {
		if (this->is_character_available_for_recruitment(hero)) {
			recruitable_heroes.push_back(hero);
		}
	}

	return recruitable_heroes;
}
/**
**  Check if the upgrade removes an existing upgrade of the player.
**
**  @param upgrade    Upgrade.
*/
bool CPlayer::UpgradeRemovesExistingUpgrade(const CUpgrade *upgrade, bool ignore_lower_priority) const
{
	for (const auto &modifier : upgrade->get_modifiers()) {
		for (size_t j = 0; j < modifier->RemoveUpgrades.size(); ++j) {
			const CUpgrade *removed_upgrade = modifier->RemoveUpgrades[j];
			bool has_upgrade = this->AiEnabled ? AiHasUpgrade(*this->Ai, removed_upgrade, true) : (UpgradeIdAllowed(*this, removed_upgrade->ID) == 'R');
			if (has_upgrade) {
				if (ignore_lower_priority && this->get_faction() != nullptr && this->get_faction()->GetUpgradePriority(removed_upgrade) < this->get_faction()->GetUpgradePriority(upgrade)) {
					continue;
				}
				return true;
			}
		}
	}
	
	return false;
}

std::string CPlayer::get_full_name() const
{
	if (!IsNetworkGame()) {
		const wyrmgus::faction *faction = this->get_faction();

		if (faction != nullptr) {
			const wyrmgus::government_type government_type = this->get_government_type();
			const wyrmgus::faction_tier tier = this->get_faction_tier();
			return faction->get_titled_name(government_type, tier);
		}
	}

	return this->Name;
}

std::string_view CPlayer::get_faction_title_name() const
{
	if (this->get_civilization() == nullptr || this->get_faction() == nullptr) {
		return wyrmgus::string::empty_str;
	}
	
	const wyrmgus::faction *faction = this->get_faction();
	const wyrmgus::government_type government_type = this->get_government_type();
	const wyrmgus::faction_tier tier = this->get_faction_tier();

	return faction->get_title_name(government_type, tier);
}

std::string_view CPlayer::GetCharacterTitleName(const character_title title_type, const wyrmgus::gender gender) const
{
	if (this->get_faction() == nullptr || title_type == wyrmgus::character_title::none || gender == wyrmgus::gender::none) {
		return wyrmgus::string::empty_str;
	}
	
	const wyrmgus::faction *faction = this->get_faction();
	const wyrmgus::government_type government_type = this->get_government_type();
	const wyrmgus::faction_tier tier = this->get_faction_tier();

	return faction->get_character_title_name(title_type, government_type, tier, gender);
}

landmass_set CPlayer::get_builder_landmasses(const wyrmgus::unit_type *building) const
{
	landmass_set builder_landmasses;

	for (const wyrmgus::unit_type *builder_type : AiHelpers.get_builders(building)) {
		if (this->GetUnitTypeAiActiveCount(builder_type) > 0) {
			std::vector<CUnit *> builder_table;

			FindPlayerUnitsByType(*this, *builder_type, builder_table, true);

			for (const CUnit *builder : builder_table) {
				const landmass *landmass = CMap::Map.get_tile_landmass(builder->tilePos, builder->MapLayer->ID);

				if (landmass == nullptr) {
					continue;
				}

				builder_landmasses.insert(landmass);
			}
		}
	}

	if (this->get_faction() != nullptr) {
		for (const wyrmgus::unit_class *builder_class : AiHelpers.get_builder_classes(building->get_unit_class())) {
			const wyrmgus::unit_type *builder_type = this->get_faction()->get_class_unit_type(builder_class);

			if (this->GetUnitTypeAiActiveCount(builder_type) > 0) {
				std::vector<CUnit *> builder_table;

				FindPlayerUnitsByType(*this, *builder_type, builder_table, true);

				for (const CUnit *builder : builder_table) {
					const landmass *landmass = CMap::Map.get_tile_landmass(builder->tilePos, builder->MapLayer->ID);

					if (landmass == nullptr) {
						continue;
					}

					builder_landmasses.insert(landmass);
				}
			}
		}
	}

	return builder_landmasses;
}

std::vector<const CUpgrade *> CPlayer::GetResearchableUpgrades()
{
	std::vector<const CUpgrade *> researchable_upgrades;

	for (const auto &kv_pair : this->UnitTypesAiActiveCount) {
		const wyrmgus::unit_type *type = kv_pair.first;

		for (const CUpgrade *upgrade : AiHelpers.get_researched_upgrades(type)) {
			if (!wyrmgus::vector::contains(researchable_upgrades, upgrade)) {
				researchable_upgrades.push_back(upgrade);
			}
		}

		for (const wyrmgus::upgrade_class *upgrade_class : AiHelpers.get_researched_upgrade_classes(type->get_unit_class())) {
			const CUpgrade *upgrade = this->get_class_upgrade(upgrade_class);
			if (upgrade != nullptr && !wyrmgus::vector::contains(researchable_upgrades, upgrade)) {
				researchable_upgrades.push_back(upgrade);
			}
		}
	}
	
	return researchable_upgrades;
}
//Wyrmgus end

/**
**  Clear all player data excepts members which don't change.
**
**  The fields that are not cleared are
**  UnitLimit, BuildingLimit, TotalUnitLimit and Allow.
*/
void CPlayer::Clear()
{
	this->Index = 0;
	this->Name.clear();
	this->Type = 0;
	this->Race = wyrmgus::defines::get()->get_neutral_civilization()->ID;
	this->Faction = -1;
	this->faction_tier = wyrmgus::faction_tier::none;
	this->government_type = wyrmgus::government_type::none;
	this->religion = nullptr;
	this->dynasty = nullptr;
	this->age = nullptr;
	this->overlord = nullptr;
	this->vassalage_type = wyrmgus::vassalage_type::none;
	this->vassals.clear();
	this->AiName.clear();
	this->Team = 0;
	this->enemies.clear();
	this->allies.clear();
	this->shared_vision.clear();
	this->StartPos.x = 0;
	this->StartPos.y = 0;
	//Wyrmgus start
	this->StartMapLayer = 0;
	//Wyrmgus end

	this->resources.clear();
	this->stored_resources.clear();
	this->max_resources.clear();
	this->last_resources.clear();
	this->incomes.clear();
	this->revenues.clear();
	//Wyrmgus start
	this->resource_demands.clear();
	this->stored_resource_demands.clear();
	//Wyrmgus end

	this->UnitTypesCount.clear();
	this->UnitTypesUnderConstructionCount.clear();
	this->UnitTypesAiActiveCount.clear();
	//Wyrmgus start
	this->Heroes.clear();
	this->Deities.clear();
	this->units_by_type.clear();
	this->units_by_class.clear();
	this->AiActiveUnitsByType.clear();
	this->last_created_unit = nullptr;

	this->available_quests.clear();
	this->current_quests.clear();
	this->completed_quests.clear();
	this->AutosellResources.clear();
	this->quest_objectives.clear();
	this->Modifiers.clear();
	//Wyrmgus end
	this->AiEnabled = false;
	this->revealed = false;
	this->Ai = 0;
	this->Units.resize(0);
	this->FreeWorkers.resize(0);
	//Wyrmgus start
	this->LevelUpUnits.resize(0);
	//Wyrmgus end
	this->NumBuildings = 0;
	//Wyrmgus start
	this->NumBuildingsUnderConstruction = 0;
	this->NumTownHalls = 0;
	//Wyrmgus end
	this->Supply = 0;
	this->Demand = 0;
	this->TradeCost = 0;
	// FIXME: can't clear limits since it's initialized already
	//	UnitLimit = 0;
	//	BuildingLimit = 0;
	//	TotalUnitLimit = 0;
	this->Score = 0;
	this->TotalUnits = 0;
	this->TotalBuildings = 0;
	this->resource_totals.clear();
	this->TotalRazings = 0;
	this->TotalKills = 0;
	//Wyrmgus start
	memset(this->UnitTypeKills, 0, sizeof(this->UnitTypeKills));
	this->LostTownHallTimer = 0;
	this->HeroCooldownTimer = 0;
	//Wyrmgus end
	this->UpgradeTimers.Clear();
	this->resource_harvest_speeds.clear();
	this->resource_return_speeds.clear();

	for (const resource *resource : resource::get_all()) {
		this->set_price(resource, resource->get_base_price());
	}

	this->SpeedBuild = CPlayer::base_speed_factor;
	this->SpeedTrain = CPlayer::base_speed_factor;
	this->SpeedUpgrade = CPlayer::base_speed_factor;
	this->SpeedResearch = CPlayer::base_speed_factor;
}


void CPlayer::AddUnit(CUnit &unit)
{
	Assert(unit.Player != this);
	Assert(unit.PlayerSlot == static_cast<size_t>(-1));
	unit.PlayerSlot = this->Units.size();
	this->Units.push_back(&unit);
	unit.Player = this;
	Assert(this->Units[unit.PlayerSlot] == &unit);
}

void CPlayer::RemoveUnit(CUnit &unit)
{
	Assert(unit.Player == this);
	//Wyrmgus start
	if (unit.PlayerSlot == -1 || this->Units[unit.PlayerSlot] != &unit) {
		log::log_error("Error in CPlayer::RemoveUnit: the unit's PlayerSlot doesn't match its position in the player's units array; Unit's PlayerSlot: " + std::to_string(unit.PlayerSlot) + ", Unit Type: \"" + (unit.Type ? unit.Type->Ident : "") + "\".");
		return;
	}
	//Wyrmgus end
	Assert(this->Units[unit.PlayerSlot] == &unit);

	//	unit.Player = nullptr; // we can remove dying unit...
	CUnit *last = this->Units.back();

	this->Units[unit.PlayerSlot] = last;
	last->PlayerSlot = unit.PlayerSlot;
	this->Units.pop_back();
	unit.PlayerSlot = static_cast<size_t>(-1);
	Assert(last == &unit || this->Units[last->PlayerSlot] == last);
}

void CPlayer::UpdateFreeWorkers()
{
	FreeWorkers.clear();
	if (FreeWorkers.capacity() != 0) {
		// Just calling FreeWorkers.clear() is not always appropriate.
		// Certain paths may leave FreeWorkers in an invalid state, so
		// it's safer to re-initialize.
		std::vector<CUnit*>().swap(FreeWorkers);
	}
	const int nunits = this->GetUnitCount();

	for (int i = 0; i < nunits; ++i) {
		CUnit &unit = this->GetUnit(i);
		//Wyrmgus start
//		if (unit.IsAlive() && unit.Type->BoolFlag[HARVESTER_INDEX].value && !unit.Removed) {
		if (unit.IsAlive() && unit.Type->BoolFlag[HARVESTER_INDEX].value && !unit.Removed && !unit.Type->BoolFlag[TRADER_INDEX].value) {
		//Wyrmgus end
			if (unit.CurrentAction() == UnitAction::Still) {
				FreeWorkers.push_back(&unit);
			}
		}
	}
}

//Wyrmgus start
void CPlayer::PerformResourceTrade()
{
	CUnit *market_unit = this->GetMarketUnit();
	
	if (!market_unit) {
		return;
	}
	
	for (size_t i = 0; i < this->AutosellResources.size(); ++i) {
		const resource *res = resource::get_all()[this->AutosellResources[i]];
		
		if ((this->get_resource(res) + this->get_stored_resource(res)) >= 100) { //sell 100 per second, as long as there is enough of the resource stored
			market_unit->sell_resource(res, this->Index);
		}
		
		//increase price due to domestic demand
		this->change_stored_resource_demand(res, this->get_effective_resource_demand(res));
		while (this->get_stored_resource_demand(res) >= 100) {
			this->increase_resource_price(res);
			this->change_stored_resource_demand(res, -100);
		}
	}
	
	for (size_t i = 0; i < LuxuryResources.size(); ++i) {
		const resource *res = resource::get_all()[LuxuryResources[i]];
		
		while ((this->get_resource(res) + this->get_stored_resource(res)) >= 100) {
			market_unit->sell_resource(res, this->Index);
		}
		
		//increase price due to domestic demand
		this->change_stored_resource_demand(res, this->get_effective_resource_demand(res));
		while (this->get_stored_resource_demand(res) >= 100) {
			this->increase_resource_price(res);
			this->change_stored_resource_demand(res, -100);
		}
	}
}

/**
**	@brief	Get whether the player has a market unit
**
**	@return	True if the player has a market unit, or false otherwise
*/
bool CPlayer::HasMarketUnit() const
{
	const int n_m = AiHelpers.SellMarkets[0].size();

	for (int i = 0; i < n_m; ++i) {
		const wyrmgus::unit_type &market_type = *AiHelpers.SellMarkets[0][i];

		if (this->GetUnitTypeCount(&market_type)) {
			return true;
		}
	}
	
	return false;
}

/**
**	@brief	Get the player's market unit, if any
**
**	@return	The market unit if present, or null otherwise
*/
CUnit *CPlayer::GetMarketUnit() const
{
	CUnit *market_unit = nullptr;
	
	const int n_m = AiHelpers.SellMarkets[0].size();

	for (int i = 0; i < n_m; ++i) {
		const wyrmgus::unit_type &market_type = *AiHelpers.SellMarkets[0][i];

		if (this->GetUnitTypeCount(&market_type)) {
			std::vector<CUnit *> market_table;
			FindPlayerUnitsByType(*this, market_type, market_table);
			if (market_table.size() > 0) {
				market_unit = market_table[SyncRand(market_table.size())];
				break;
			}
		}
	}
	
	return market_unit;
}

std::vector<int> CPlayer::GetAutosellResources() const
{
	return this->AutosellResources;
}

void CPlayer::AutosellResource(const int resource)
{
	if (std::find(this->AutosellResources.begin(), this->AutosellResources.end(), resource) != this->AutosellResources.end()) {
		this->AutosellResources.erase(std::remove(this->AutosellResources.begin(), this->AutosellResources.end(), resource), this->AutosellResources.end());
	} else {
		this->AutosellResources.push_back(resource);
	}
}

void CPlayer::UpdateLevelUpUnits()
{
	LevelUpUnits.clear();
	if (LevelUpUnits.capacity() != 0) {
		// Just calling LevelUpUnits.clear() is not always appropriate.
		// Certain paths may leave LevelUpUnits in an invalid state, so
		// it's safer to re-initialize.
		std::vector<CUnit*>().swap(LevelUpUnits);
	}
	const int nunits = this->GetUnitCount();

	for (int i = 0; i < nunits; ++i) {
		CUnit &unit = this->GetUnit(i);
		if (unit.IsAlive() && unit.Variable[LEVELUP_INDEX].Value >= 1) {
			LevelUpUnits.push_back(&unit);
		}
	}
}

void CPlayer::update_quest_pool()
{
	if (wyrmgus::game::get()->get_current_campaign() == nullptr) { // in-game quests only while playing the campaign mode
		return;
	}

	if (this->get_faction() == nullptr) {
		return;
	}
	
	const bool exausted_available_quests = this->available_quests.empty();
	
	this->available_quests.clear();
	
	std::vector<wyrmgus::quest *> potential_quests;
	for (wyrmgus::quest *quest : wyrmgus::quest::get_all()) {
		if (this->can_quest_be_available(quest)) {
			potential_quests.push_back(quest);
		}
	}
	
	for (int i = 0; i < CPlayer::max_quest_pool; ++i) { //fill the quest pool with quests up to the max quantity
		if (potential_quests.empty()) {
			break;
		}

		wyrmgus::quest *quest = wyrmgus::vector::take_random(potential_quests);
		this->available_quests.push_back(quest);
	}

	this->on_available_quests_changed();

	// notify the player when new quests are available (but only if the player has already exausted the quests available to him, so that they aren't bothered if they choose not to engage with the quest system)
	if (this == CPlayer::GetThisPlayer() && GameCycle >= CYCLES_PER_MINUTE && this->available_quests.size() > 0 && exausted_available_quests && this->NumTownHalls > 0 && this->get_current_quests().size() < CPlayer::max_current_quests) {
		CPlayer::GetThisPlayer()->Notify("%s", _("New quests available"));
	}
	
	if (this->AiEnabled && this->NumTownHalls > 0) { // if is an AI player, accept all quests that it can
		int available_quest_quantity = this->available_quests.size();
		for (int i = (available_quest_quantity  - 1); i >= 0; --i) {
			if (this->get_current_quests().size() >= CPlayer::max_current_quests) {
				break;
			}

			if (this->can_accept_quest(this->available_quests[i])) { // something may have changed, so recheck if the player is able to accept the quest
				this->accept_quest(this->available_quests[i]);
			}
		}
	}
}

void CPlayer::on_available_quests_changed()
{
	if (this == CPlayer::GetThisPlayer()) {
		for (wyrmgus::button *button : wyrmgus::button::get_all()) {
			if (button->Action != ButtonCmd::Quest || button->Value >= static_cast<int>(this->available_quests.size())) {
				continue;
			}
			
			const wyrmgus::quest *quest = this->available_quests[button->Value];
			button->Hint = "Quest: " + quest->get_name();
			button->Description = quest->get_description() + "\n \nObjectives:";
			for (const auto &objective : quest->get_objectives()) {
				button->Description += "\n- ";
				if (!objective->get_objective_string().empty()) {
					button->Description += objective->get_objective_string();
				} else {
					button->Description += objective->generate_objective_string(this);
				}
			}
			for (const std::string &objective_string : quest->get_objective_strings()) {
				button->Description += "\n" + objective_string;
			}
			const std::string rewards_string = quest->get_rewards_string(this);
			if (!rewards_string.empty()) {
				button->Description += "\n \nRewards:\n" + rewards_string;
			}
			if (!quest->get_hint().empty()) {
				button->Description += "\n \nHint: " + quest->get_hint();
			}
			if (quest->HighestCompletedDifficulty > DifficultyNoDifficulty) {
				std::string highest_completed_difficulty;
				if (quest->HighestCompletedDifficulty == DifficultyEasy) {
					highest_completed_difficulty = "Easy";
				} else if (quest->HighestCompletedDifficulty == DifficultyNormal) {
					highest_completed_difficulty = "Normal";
				} else if (quest->HighestCompletedDifficulty == DifficultyHard) {
					highest_completed_difficulty = "Hard";
				} else if (quest->HighestCompletedDifficulty == DifficultyBrutal) {
					highest_completed_difficulty = "Brutal";
				}
				button->Description += "\n \nHighest Completed Difficulty: " + highest_completed_difficulty;
			}
			
		}
		
		if (!Selected.empty() && Selected[0]->Type->BoolFlag[TOWNHALL_INDEX].value) {
			UI.ButtonPanel.Update();
		}
	}
}

void CPlayer::update_current_quests()
{
	for (const auto &objective : this->get_quest_objectives()) {
		objective->update_counter();
	}
	
	for (int i = (this->current_quests.size()  - 1); i >= 0; --i) {
		wyrmgus::quest *quest = this->current_quests[i];
		const std::pair<bool, std::string> failure_result = this->check_quest_failure(quest);
		if (failure_result.first) {
			this->fail_quest(quest, failure_result.second);
		} else if (this->check_quest_completion(quest)) {
			this->complete_quest(quest);
		}
	}
}

void CPlayer::accept_quest(wyrmgus::quest *quest)
{
	if (!quest) {
		return;
	}
	
	wyrmgus::vector::remove(this->available_quests, quest);
	this->current_quests.push_back(quest);
	
	for (const auto &quest_objective : quest->get_objectives()) {
		auto objective = std::make_unique<wyrmgus::player_quest_objective>(quest_objective.get(), this);
		this->quest_objectives.push_back(std::move(objective));
	}
	
	CclCommand("trigger_player = " + std::to_string(this->Index) + ";");
	
	if (quest->AcceptEffects) {
		quest->AcceptEffects->pushPreamble();
		quest->AcceptEffects->run();
	}

	if (quest->get_accept_effects() != nullptr) {
		wyrmgus::context ctx;
		ctx.current_player = this;
		quest->get_accept_effects()->do_effects(this, ctx);
	}

	this->on_available_quests_changed();
	
	this->update_current_quests();
}

void CPlayer::complete_quest(wyrmgus::quest *quest)
{
	if (wyrmgus::vector::contains(this->completed_quests, quest)) {
		return;
	}
	
	this->remove_current_quest(quest);
	
	this->completed_quests.push_back(quest);
	if (quest->is_competitive()) {
		quest->CurrentCompleted = true;
	}
	
	CclCommand("trigger_player = " + std::to_string(this->Index) + ";");
	
	if (quest->CompletionEffects) {
		quest->CompletionEffects->pushPreamble();
		quest->CompletionEffects->run();
	}

	if (quest->get_completion_effects() != nullptr) {
		wyrmgus::context ctx;
		ctx.current_player = this;
		quest->get_completion_effects()->do_effects(this, ctx);
	}
	
	if (this == CPlayer::GetThisPlayer()) {
		SetQuestCompleted(quest->get_identifier(), GameSettings.Difficulty);
		SaveQuestCompletion();

		const wyrmgus::campaign *current_campaign = wyrmgus::game::get()->get_current_campaign();
		if (current_campaign != nullptr && current_campaign->get_quest() == quest) {
			wyrmgus::context ctx;
			ctx.current_player = this;
			wyrmgus::defines::get()->get_campaign_victory_dialogue()->call(this, ctx);
		}

		std::string rewards_string = quest->get_rewards_string(this);
		if (!rewards_string.empty()) {
			wyrmgus::string::replace(rewards_string, "\n", "\\n");
			wyrmgus::string::replace(rewards_string, "\t", "\\t");
			rewards_string = "Rewards:\\n" + rewards_string;
		}
		CclCommand("if (GenericDialog ~= nil) then GenericDialog(\"Quest Completed\", \"You have completed the " + quest->get_name() + " quest!\\n\\n" + rewards_string + "\", nil, \"" + (quest->get_icon() ? quest->get_icon()->get_identifier() : "") + "\", \"" + (quest->get_player_color() ? quest->get_player_color()->get_identifier() : "") + "\", " + std::to_string(quest->get_icon() ? quest->get_icon()->get_frame() : 0) + ") end;");
	}
}

void CPlayer::fail_quest(wyrmgus::quest *quest, const std::string &fail_reason)
{
	this->remove_current_quest(quest);
	
	CclCommand("trigger_player = " + std::to_string(this->Index) + ";");
	
	if (quest->FailEffects) {
		quest->FailEffects->pushPreamble();
		quest->FailEffects->run();
	}
	
	if (quest->get_failure_effects() != nullptr) {
		wyrmgus::context ctx;
		ctx.current_player = this;
		quest->get_failure_effects()->do_effects(this, ctx);
	}

	if (this == CPlayer::GetThisPlayer()) {
		const wyrmgus::campaign *current_campaign = wyrmgus::game::get()->get_current_campaign();
		if (current_campaign != nullptr && current_campaign->get_quest() == quest) {
			wyrmgus::context ctx;
			ctx.current_player = this;
			wyrmgus::defines::get()->get_campaign_defeat_dialogue()->call(this, ctx);
		}

		CclCommand("if (GenericDialog ~= nil) then GenericDialog(\"Quest Failed\", \"You have failed the " + quest->get_name() + " quest! " + fail_reason + "\", nil, \"" + (quest->get_icon() ? quest->get_icon()->get_identifier() : "") + "\", \"" + (quest->get_player_color() ? quest->get_player_color()->get_identifier() : "") + "\", " + std::to_string(quest->get_icon() ? quest->get_icon()->get_frame() : 0) + ") end;");
	}
}

void CPlayer::remove_current_quest(wyrmgus::quest *quest)
{
	wyrmgus::vector::remove(this->current_quests, quest);
	
	for (int i = (this->quest_objectives.size()  - 1); i >= 0; --i) {
		if (this->quest_objectives[i]->get_quest_objective()->get_quest() == quest) {
			wyrmgus::vector::remove(this->quest_objectives, this->quest_objectives[i]);
		}
	}
}

bool CPlayer::can_quest_be_available(const wyrmgus::quest *quest) const
{
	if (quest->Hidden || quest->is_unobtainable()) {
		return false;
	}

	return this->can_accept_quest(quest);
}

bool CPlayer::can_accept_quest(const wyrmgus::quest *quest) const
{
	if (this->get_faction() == nullptr) {
		return false;
	}

	if (quest->CurrentCompleted) {
		return false;
	}
	
	if (vector::contains(this->current_quests, quest) || vector::contains(this->completed_quests, quest)) {
		return false;
	}

	if (quest->get_conditions() != nullptr && !quest->get_conditions()->check(this)) {
		return false;
	}
	
	int recruit_heroes_quantity = 0;
	for (const auto &objective : quest->get_objectives()) {
		if (!objective->is_quest_acceptance_allowed(this)) {
			return false;
		}

		if (objective->get_objective_type() == wyrmgus::objective_type::recruit_hero) {
			++recruit_heroes_quantity;
		}
	}
	
	if (recruit_heroes_quantity > 0 && (this->Heroes.size() + recruit_heroes_quantity) > PlayerHeroMax) {
		return false;
	}

	for (const wyrmgus::quest *current_quest : this->current_quests) {
		if (quest->overlaps_with(current_quest)) {
			return false;
		}
	}
	
	for (const wyrmgus::character *hero : quest->HeroesMustSurvive) {
		if (!this->HasHero(hero)) {
			return false;
		}
	}

	if (quest->Conditions != nullptr) {
		CclCommand("trigger_player = " + std::to_string(this->Index) + ";");
		quest->Conditions->pushPreamble();
		quest->Conditions->run(1);
		return quest->Conditions->popBoolean();
	} else {
		return true;
	}
}

bool CPlayer::check_quest_completion(const wyrmgus::quest *quest) const
{
	if (quest->is_uncompleteable()) {
		return false;
	}
	
	for (const auto &objective : this->get_quest_objectives()) {
		const wyrmgus::quest_objective *quest_objective = objective->get_quest_objective();
		if (quest_objective->get_quest() != quest) {
			continue;
		}
		if (quest_objective->get_quantity() != 0 && objective->get_counter() < quest_objective->get_quantity()) {
			return false;
		}
	}
	
	return true;
}

//returns the reason for failure (empty if none)
std::pair<bool, std::string> CPlayer::check_quest_failure(const wyrmgus::quest *quest) const
{
	if (quest->CurrentCompleted) { // quest already completed by someone else
		return std::make_pair(true, "Another faction has completed the quest before you could.");
	}

	for (const auto &objective : this->get_quest_objectives()) {
		const wyrmgus::quest_objective *quest_objective = objective->get_quest_objective();
		if (quest_objective->get_quest() != quest) {
			continue;
		}

		//"unfailable" quests should also fail when a hero which should survive dies
		if (quest->is_unfailable() && quest_objective->get_objective_type() != wyrmgus::objective_type::hero_must_survive) {
			continue;
		}

		if (quest_objective->get_quantity() != 0 && objective->get_counter() >= quest_objective->get_quantity()) {
			//objective already fulfilled
			continue;
		}

		std::pair<bool, std::string> result = quest_objective->check_failure(this);
		if (result.first) {
			return result;
		}
	}

	//unfailable" quests should also fail when a hero which should survive dies
	for (const wyrmgus::character *character : quest->HeroesMustSurvive) {
		if (!this->HasHero(character)) {
			return std::make_pair(true, "A hero necessary for the quest has died.");
		}
	}

	return std::make_pair(false, std::string());
}

bool CPlayer::has_quest(const wyrmgus::quest *quest) const
{
	return wyrmgus::vector::contains(this->current_quests, quest);
}

bool CPlayer::is_quest_completed(const wyrmgus::quest *quest) const
{
	return wyrmgus::vector::contains(this->completed_quests, quest);
}

void CPlayer::on_unit_built(const CUnit *unit)
{
	bool counter_changed = false;

	for (const auto &objective : this->get_quest_objectives()) {
		const int old_counter = objective->get_counter();
		objective->on_unit_built(unit);
		const int new_counter = objective->get_counter();

		if (old_counter != new_counter) {
			counter_changed = true;
		}
	}

	if (counter_changed) {
		this->update_current_quests();
	}
}

void CPlayer::on_unit_destroyed(const CUnit *unit)
{
	bool counter_changed = false;

	for (const auto &objective : this->get_quest_objectives()) {
		const int old_counter = objective->get_counter();
		objective->on_unit_destroyed(unit);
		const int new_counter = objective->get_counter();

		if (old_counter != new_counter) {
			counter_changed = true;
		}
	}

	if (counter_changed) {
		this->update_current_quests();
	}
}

void CPlayer::on_resource_gathered(const wyrmgus::resource *resource, const int quantity)
{
	bool counter_changed = false;

	for (const auto &objective : this->get_quest_objectives()) {
		const int old_counter = objective->get_counter();
		objective->on_resource_gathered(resource, quantity);
		const int new_counter = objective->get_counter();

		if (old_counter != new_counter) {
			counter_changed = true;
		}
	}

	if (counter_changed) {
		this->update_current_quests();
	}
}

void CPlayer::AddModifier(CUpgrade *modifier, int cycles)
{
	if (this->Allow.Upgrades[modifier->ID] == 'R') {
		for (size_t i = 0; i < this->Modifiers.size(); ++i) { //if already has the modifier, make it have the greater duration of the new or old one
			if (this->Modifiers[i].first == modifier) {
				this->Modifiers[i].second = std::max(this->Modifiers[i].second, (int) (GameCycle + cycles));
			}
		}
	} else {
		this->Modifiers.push_back(std::pair<CUpgrade *, int>(modifier, GameCycle + cycles));
		UpgradeAcquire(*this, modifier);
	}
	
}

void CPlayer::RemoveModifier(CUpgrade *modifier)
{
	if (this->Allow.Upgrades[modifier->ID] == 'R') {
		UpgradeLost(*this, modifier->ID);
		for (size_t i = 0; i < this->Modifiers.size(); ++i) { //if already has the modifier, make it have the greater duration of the new or old one
			if (this->Modifiers[i].first == modifier) {
				this->Modifiers.erase(std::remove(this->Modifiers.begin(), this->Modifiers.end(), this->Modifiers[i]), this->Modifiers.end());
				break;
			}
		}
	}
}

bool CPlayer::at_war() const
{
	for (int i = 0; i < PlayerNumNeutral; ++i) {
		const CPlayer *other_player = CPlayer::Players[i];

		if (other_player == this) {
			continue;
		}

		if (!this->HasContactWith(*other_player)) {
			continue;
		}

		if (!other_player->is_alive()) {
			continue;
		}

		if (this->IsEnemy(*other_player)) {
			return true;
		}
	}
	
	return false;
}
//Wyrmgus end

std::vector<CUnit *>::const_iterator CPlayer::UnitBegin() const
{
	return Units.begin();
}

std::vector<CUnit *>::iterator CPlayer::UnitBegin()
{
	return Units.begin();
}

std::vector<CUnit *>::const_iterator CPlayer::UnitEnd() const
{
	return Units.end();
}

std::vector<CUnit *>::iterator CPlayer::UnitEnd()
{
	return Units.end();
}

CUnit &CPlayer::GetUnit(int index) const
{
	return *Units[index];
}

int CPlayer::GetUnitCount() const
{
	return static_cast<int>(Units.size());
}



/*----------------------------------------------------------------------------
--  Resource management
----------------------------------------------------------------------------*/

/**
**  Gets the player resource.
**
**  @param resource  Resource to get.
**  @param type      Storing type
**
**  @note Storing types: 0 - overall store, 1 - store buildings, 2 - both
*/
int CPlayer::get_resource(const wyrmgus::resource *resource, const resource_storage_type type) const
{
	switch (type) {
		case resource_storage_type::overall:
			return this->get_resource(resource);
		case resource_storage_type::building:
			return this->get_stored_resource(resource);
		case resource_storage_type::both:
			return this->get_resource(resource) + this->get_stored_resource(resource);
		default:
			DebugPrint("Wrong resource type\n");
			return -1;
	}
}

/**
**  Adds/subtracts some resources to/from the player store
**
**  @param resource  Resource to add/subtract.
**  @param value     How many of this resource (can be negative).
**  @param store     If true, sets the building store resources, else the overall resources.
*/
void CPlayer::change_resource(const wyrmgus::resource *resource, const int value, const bool store)
{
	if (value < 0) {
		const int fromStore = std::min(this->get_stored_resource(resource), abs(value));
		this->change_stored_resource(resource, -fromStore);
		this->change_resource(resource, -(abs(value) - fromStore));
		this->set_resource(resource, std::max(this->get_resource(resource), 0));
	} else {
		if (store && this->get_max_resource(resource) != -1) {
			this->change_stored_resource(resource,  std::min(value, this->get_max_resource(resource) - this->get_stored_resource(resource)));
		} else {
			this->change_resource(resource,  value);
		}
	}
}

/**
**  Change the player resource.
**
**  @param resource  Resource to change.
**  @param value     How many of this resource.
**  @param type      Resource types: 0 - overall store, 1 - store buildings, 2 - both
*/
void CPlayer::set_resource(const wyrmgus::resource *resource, const int value, const resource_storage_type type)
{
	switch (type) {
		case resource_storage_type::both:
			if (this->get_max_resource(resource) != -1) {
				const int to_res = std::max(0, value - this->get_stored_resource(resource));
				this->set_resource(resource, std::max(0, to_res));
				this->set_stored_resource(resource, std::min(value - to_res, this->get_max_resource(resource)));
			} else {
				this->set_resource(resource, std::max(0, value));
			}
			break;
		case resource_storage_type::building:
			if (this->get_max_resource(resource) != -1) {
				this->set_stored_resource(resource, std::min(value, this->get_max_resource(resource)));
			}
			break;
		case resource_storage_type::overall:
			this->set_resource(resource, std::max(0, value));
			break;
		default:
			break;
	}
}

/**
**  Check, if there enough resources for action.
**
**  @param resource  Resource to change.
**  @param value     How many of this resource.
*/
bool CPlayer::check_resource(const resource *resource, const int value)
{
	int result = this->get_resource(resource);
	if (this->get_max_resource(resource) != -1) {
		result += this->get_stored_resource(resource);
	}
	return result < value ? false : true;
}

void CPlayer::increase_resource_price(const resource *resource)
{
	int price_change = resource->get_base_price() / std::max(this->get_price(resource), 100);
	price_change = std::max(1, price_change);
	this->change_price(resource, price_change);
}

void CPlayer::decrease_resource_price(const resource *resource)
{
	int price_change = this->get_price(resource) / resource->get_base_price();
	price_change = std::max(1, price_change);
	this->change_price(resource, -price_change);
	this->set_price(resource, std::max(1, this->get_price(resource)));
}

/**
**  Converges prices with another player (and returns how many convergences were effected)
*/
int CPlayer::ConvergePricesWith(CPlayer &player, int max_convergences)
{
	int convergences = 0;
	
	bool converged = true;
	while (converged) {
		converged = false;

		for (const resource *resource : resource::get_all()) {
			if (resource == defines::get()->get_time_resource()) {
				continue;
			}

			if (resource->get_base_price() == 0) {
				continue;
			}

			int convergence_increase = 100;

			if (this->get_price(resource) < player.get_price(resource) && convergences < max_convergences) {
				this->increase_resource_price(resource);
				convergences += convergence_increase;
				converged = true;

				if (this->get_price(resource) < player.get_price(resource) && convergences < max_convergences) {
					//now do the convergence for the other side as well, if possible
					player.decrease_resource_price(resource);
					convergences += convergence_increase;
					converged = true;
				}
			} else if (this->get_price(resource) > player.get_price(resource) && convergences < max_convergences) {
				this->decrease_resource_price(resource);
				convergences += convergence_increase;
				converged = true;

				if (this->get_price(resource) > player.get_price(resource) && convergences < max_convergences) { //do the convergence for the other side as well, if possible
					player.increase_resource_price(resource);
					convergences += convergence_increase;
					converged = true;
				}
			}
		}
	}
	
	return convergences;
}

int CPlayer::get_resource_price(const resource *resource) const
{
	if (resource == defines::get()->get_wealth_resource()) {
		return 100;
	}
	
	return this->get_price(resource);
}

/**
**  Get the effective resource demand for the player, given the current prices
**
**  @param resource  Resource.
*/
int CPlayer::get_effective_resource_demand(const resource *resource) const
{
	int resource_demand = this->get_resource_demand(resource);
	
	if (this->get_price(resource) != 0) {
		resource_demand *= resource->get_base_price();
		resource_demand /= this->get_price(resource);
	}
	
	if (resource->DemandElasticity != 100) {
		resource_demand = this->get_resource_demand(resource) + ((resource_demand - this->get_resource_demand(resource)) * resource->DemandElasticity / 100);
	}
	
	resource_demand = std::max(resource_demand, 0);

	return resource_demand;
}

/**
**  Get the effective sell price of a resource
*/
int CPlayer::get_effective_resource_sell_price(const resource *resource, const int traded_quantity) const
{
	if (resource == defines::get()->get_wealth_resource()) {
		return 100;
	}
	
	int price = traded_quantity * this->get_price(resource) / 100 * (100 - this->TradeCost) / 100;
	price = std::max(1, price);
	return price;
}

/**
**  Get the effective buy quantity of a resource
*/
int CPlayer::get_effective_resource_buy_price(const resource *resource, const int traded_quantity) const
{
	int price = traded_quantity * this->get_price(resource) / 100 * 100 / (100 - this->TradeCost);
	price = std::max(1, price);
	return price;
}

/**
**  Get the total price difference between this player and another one
*/
int CPlayer::GetTotalPriceDifferenceWith(const CPlayer &player) const
{
	int difference = 0;

	for (const resource *resource : resource::get_all()) {
		if (resource == defines::get()->get_time_resource()) {
			continue;
		}

		if (resource->get_base_price() == 0) {
			continue;
		}

		difference += abs(this->get_price(resource) - player.get_price(resource));
	}

	return difference;
}

/**
**  Get the trade potential between this player and another one
*/
int CPlayer::GetTradePotentialWith(const CPlayer &player) const
{
	int trade_potential = 0;

	for (const resource *resource : resource::get_all()) {
		if (resource == defines::get()->get_time_resource()) {
			continue;
		}

		if (resource->get_base_price() == 0) {
			continue;
		}

		const int price_difference = abs(this->get_price(resource) - player.get_price(resource));
		trade_potential += price_difference * 100;
	}

	trade_potential = std::max(trade_potential, 10);
	
	return trade_potential;
}

void CPlayer::pay_overlord_tax(const wyrmgus::resource *resource, const int taxable_quantity)
{
	if (this->get_overlord() == nullptr) {
		return;
	}

	//if the player has an overlord, give 10% of the resources gathered to them
	const int quantity = taxable_quantity / 10;

	if (quantity == 0) {
		return;
	}

	this->get_overlord()->change_resource(resource, quantity, true);
	this->get_overlord()->change_resource_total(resource, quantity);
	this->change_resource(resource, -quantity, true);

	//make the overlord pay tax to their overlord in turn (if they have one)
	this->get_overlord()->pay_overlord_tax(resource, quantity);
}

int CPlayer::GetUnitTotalCount(const wyrmgus::unit_type &type) const
{
	int count = this->GetUnitTypeCount(&type);
	for (std::vector<CUnit *>::const_iterator it = this->UnitBegin(); it != this->UnitEnd(); ++it) {
		//Wyrmgus start
		if (*it == nullptr) {
			fprintf(stderr, "Error in CPlayer::GetUnitTotalCount: unit of player %d is null.\n", this->Index);
			continue;
		}
		//Wyrmgus end
		CUnit &unit = **it;

		if (unit.CurrentAction() == UnitAction::UpgradeTo) {
			COrder_UpgradeTo &order = dynamic_cast<COrder_UpgradeTo &>(*unit.CurrentOrder());
			if (order.GetUnitType().Slot == type.Slot) {
				++count;
			}
		}
	}
	return count;
}

/**
**  Check if the unit-type didn't break any unit limits.
**
**  @param type    Type of unit.
**
**  @return        True if enough, negative on problem.
**
**  @note The return values of the PlayerCheck functions are inconsistent.
*/
int CPlayer::CheckLimits(const wyrmgus::unit_type &type) const
{
	//  Check game limits.
	if (type.BoolFlag[BUILDING_INDEX].value && NumBuildings >= BuildingLimit) {
		Notify("%s", _("Building Limit Reached"));
		return -1;
	}
	if (!type.BoolFlag[BUILDING_INDEX].value && (this->GetUnitCount() - NumBuildings) >= UnitLimit) {
		Notify("%s", _("Unit Limit Reached"));
		return -2;
	}
	//Wyrmgus start
//	if (this->Demand + type.Stats[this->Index].Variables[DEMAND_INDEX].Value > this->Supply && type.Stats[this->Index].Variables[DEMAND_INDEX].Value) {
	if (this->Demand + (type.Stats[this->Index].Variables[DEMAND_INDEX].Value * (type.TrainQuantity ? type.TrainQuantity : 1)) > this->Supply && type.Stats[this->Index].Variables[DEMAND_INDEX].Value) {
	//Wyrmgus end
		//Wyrmgus start
//		Notify("%s", _("Insufficient Supply, increase Supply."));
		Notify("%s", _("Insufficient Food Supply, increase Food Supply."));
		//Wyrmgus end
		return -3;
	}
	if (this->GetUnitCount() >= TotalUnitLimit) {
		Notify("%s", _("Total Unit Limit Reached"));
		return -4;
	}
	if (GetUnitTotalCount(type) >= Allow.Units[type.Slot]) {
		Notify(_("Limit of %d reached for this unit type"), Allow.Units[type.Slot]);
		return -6;
	}
	return 1;
}

/**
**  Check if enough resources for are available.
**
**  @param costs   How many costs.
**
**  @return        False if all enough, otherwise a bit mask.
**
**  @note The return values of the PlayerCheck functions are inconsistent.
*/
int CPlayer::CheckCosts(const resource_map<int> &costs, const bool notify) const
{
	bool sound_played = false;
	int err = 0;

	for (const auto &[resource, cost] : costs) {
		if (resource == defines::get()->get_time_resource()) {
			continue;
		}

		const int resource_index = resource->get_index();

		if (this->get_resource(resource) + this->get_stored_resource(resource) >= cost) {
			continue;
		}

		if (notify) {
			const std::string &name = resource->get_identifier();
			const std::string &action_name = resource->get_action_name();

			Notify(_("Not enough %s... %s more %s."), _(name.c_str()), _(action_name.c_str()), _(name.c_str()));

			if (this == CPlayer::GetThisPlayer() && this->get_civilization() != nullptr && !sound_played) {
				const wyrmgus::sound *sound = this->get_civilization()->get_not_enough_resource_sound(resource);
				if (sound != nullptr) {
					sound_played = true;
					PlayGameSound(sound, MaxSampleVolume);
				}
			}
		}

		err |= 1 << resource_index;
	}

	return err;
}

/**
**  Check if enough resources for new unit is available.
**
**  @param type    Type of unit.
**
**  @return        False if all enough, otherwise a bit mask.
*/
int CPlayer::CheckUnitType(const wyrmgus::unit_type &type, bool hire) const
{
	//Wyrmgus start
//	return this->CheckCosts(type.Stats[this->Index].Costs);
	const resource_map<int> type_costs = this->GetUnitTypeCosts(&type, hire);
	return this->CheckCosts(type_costs);
	//Wyrmgus end
}

/**
**  Add costs to the resources
**
**  @param costs   How many costs.
*/
void CPlayer::AddCosts(const int *costs)
{
	for (int i = 1; i < MaxCosts; ++i) {
		change_resource(wyrmgus::resource::get_all()[i], costs[i], false);
	}
}

/**
**  Add the costs of an unit type to resources
**
**  @param type    Type of unit.
*/
void CPlayer::AddUnitType(const wyrmgus::unit_type &type, bool hire)
{
	//Wyrmgus start
//	AddCosts(type.Stats[this->Index].Costs);
	const resource_map<int> type_costs = this->GetUnitTypeCosts(&type, hire);
	AddCostsFactor(type_costs, 100);
	//Wyrmgus end
}

/**
**  Add a factor of costs to the resources
**
**  @param costs   How many costs.
**  @param factor  Factor of the costs to apply.
*/
void CPlayer::AddCostsFactor(const resource_map<int> &costs, const int factor)
{
	if (factor == 0) {
		return;
	}
	
	for (const auto &[resource, cost] : costs) {
		if (resource == defines::get()->get_time_resource()) {
			continue;
		}

		change_resource(resource, cost * factor / 100, true);
	}
}

void CPlayer::subtract_costs(const resource_map<int> &costs)
{
	for (const auto &[resource, cost] : costs) {
		if (resource == defines::get()->get_time_resource()) {
			continue;
		}

		this->change_resource(resource, -cost, true);
	}
}

/**
**  Subtract the costs of new unit from resources
**
**  @param type    Type of unit.
*/
void CPlayer::SubUnitType(const wyrmgus::unit_type &type, bool hire)
{
	//Wyrmgus start
//	this->SubCosts(type.Stats[this->Index].Costs);
	const resource_map<int> type_costs = this->GetUnitTypeCosts(&type, hire);
	this->SubCostsFactor(type_costs, 100);
	//Wyrmgus end
}

/**
**  Subtract a factor of costs from the resources
**
**  @param costs   How many costs.
**  @param factor  Factor of the costs to apply.
*/
void CPlayer::SubCostsFactor(const resource_map<int> &costs, const int factor)
{
	for (const auto &[resource, cost] : costs) {
		if (resource == defines::get()->get_time_resource()) {
			continue;
		}

		this->change_resource(resource, -cost * 100 / factor, false);
	}
}

//Wyrmgus start
/**
**  Gives the cost of a unit type for the player
*/
resource_map<int> CPlayer::GetUnitTypeCosts(const unit_type *type, const bool hire, const bool ignore_one) const
{
	resource_map<int> costs;

	if (hire) {
		costs[defines::get()->get_wealth_resource()] = type->Stats[this->Index].get_price();
	} else {
		costs = type->Stats[this->Index].get_costs();
	}

	for (auto &[resource, cost] : costs) {
		if (type->TrainQuantity != 0) {
			cost *= type->TrainQuantity;
		}

		if (type->CostModifier != 0) {
			int type_count = this->GetUnitTypeCount(type) + this->GetUnitTypeUnderConstructionCount(type);
			if (ignore_one) {
				type_count--;
			}

			for (int j = 0; j < type_count; ++j) {
				cost *= 100 + type->CostModifier;
				cost /= 100;
			}
		}
	}

	return costs;
}

int CPlayer::GetUnitTypeCostsMask(const wyrmgus::unit_type *type, bool hire) const
{
	int costs_mask = 0;
	
	const resource_map<int> type_costs = this->GetUnitTypeCosts(type, hire);
	
	for (const auto &[resource, cost] : type_costs) {
		if (resource == defines::get()->get_time_resource()) {
			continue;
		}

		if (cost > 0) {
			costs_mask |= 1 << resource->get_index();
		}
	}
	
	return costs_mask;
}

/**
**  Gives the cost of an upgrade for the player
*/
resource_map<int> CPlayer::GetUpgradeCosts(const CUpgrade *upgrade) const
{
	resource_map<int> costs = upgrade->get_costs();

	for (const auto &[resource, scaled_cost] : upgrade->get_scaled_costs()) {
		for (const wyrmgus::unit_type *unit_type : upgrade->get_scaled_cost_unit_types()) {
			costs[resource] += scaled_cost * this->GetUnitTypeCount(unit_type);
		}

		for (const wyrmgus::unit_class *unit_class : upgrade->get_scaled_cost_unit_classes()) {
			costs[resource] += scaled_cost * this->get_unit_class_count(unit_class);
		}
	}

	return costs;
}

int CPlayer::GetUpgradeCostsMask(const CUpgrade *upgrade) const
{
	int costs_mask = 0;
	
	const resource_map<int> upgrade_costs = AiPlayer->Player->GetUpgradeCosts(upgrade);
	
	for (const auto &[resource, cost] : upgrade_costs) {
		if (resource == defines::get()->get_time_resource()) {
			continue;
		}

		if (cost > 0) {
			costs_mask |= 1 << resource->get_index();
		}
	}
	
	return costs_mask;
}

//Wyrmgus end

void CPlayer::SetUnitTypeCount(const wyrmgus::unit_type *type, int quantity)
{
	if (!type) {
		return;
	}
	
	if (quantity <= 0) {
		if (this->UnitTypesCount.find(type) != this->UnitTypesCount.end()) {
			this->UnitTypesCount.erase(type);
		}
	} else {
		this->UnitTypesCount[type] = quantity;
	}
}

void CPlayer::ChangeUnitTypeCount(const wyrmgus::unit_type *type, int quantity)
{
	this->SetUnitTypeCount(type, this->GetUnitTypeCount(type) + quantity);
}

int CPlayer::GetUnitTypeCount(const wyrmgus::unit_type *type) const
{
	if (type != nullptr && this->UnitTypesCount.find(type) != this->UnitTypesCount.end()) {
		return this->UnitTypesCount.find(type)->second;
	} else {
		return 0;
	}
}

void CPlayer::SetUnitTypeUnderConstructionCount(const wyrmgus::unit_type *type, int quantity)
{
	if (!type) {
		return;
	}
	
	if (quantity <= 0) {
		if (this->UnitTypesUnderConstructionCount.find(type) != this->UnitTypesUnderConstructionCount.end()) {
			this->UnitTypesUnderConstructionCount.erase(type);
		}
	} else {
		this->UnitTypesUnderConstructionCount[type] = quantity;
	}
}

void CPlayer::ChangeUnitTypeUnderConstructionCount(const wyrmgus::unit_type *type, int quantity)
{
	this->SetUnitTypeUnderConstructionCount(type, this->GetUnitTypeUnderConstructionCount(type) + quantity);
}

int CPlayer::GetUnitTypeUnderConstructionCount(const wyrmgus::unit_type *type) const
{
	if (type && this->UnitTypesUnderConstructionCount.find(type) != this->UnitTypesUnderConstructionCount.end()) {
		return this->UnitTypesUnderConstructionCount.find(type)->second;
	} else {
		return 0;
	}
}

void CPlayer::SetUnitTypeAiActiveCount(const wyrmgus::unit_type *type, int quantity)
{
	if (!type) {
		return;
	}
	
	if (quantity <= 0) {
		if (this->UnitTypesAiActiveCount.find(type) != this->UnitTypesAiActiveCount.end()) {
			this->UnitTypesAiActiveCount.erase(type);
		}
	} else {
		this->UnitTypesAiActiveCount[type] = quantity;
	}
}

void CPlayer::ChangeUnitTypeAiActiveCount(const wyrmgus::unit_type *type, int quantity)
{
	this->SetUnitTypeAiActiveCount(type, this->GetUnitTypeAiActiveCount(type) + quantity);
}

int CPlayer::GetUnitTypeAiActiveCount(const wyrmgus::unit_type *type) const
{
	if (type && this->UnitTypesAiActiveCount.find(type) != this->UnitTypesAiActiveCount.end()) {
		return this->UnitTypesAiActiveCount.find(type)->second;
	} else {
		return 0;
	}
}

void CPlayer::IncreaseCountsForUnit(CUnit *unit, const bool type_change)
{
	const wyrmgus::unit_type *type = unit->Type;

	this->ChangeUnitTypeCount(type, 1);
	this->units_by_type[type].push_back(unit);

	if (type->get_unit_class() != nullptr) {
		this->units_by_class[type->get_unit_class()].push_back(unit);
	}
	
	if (unit->Active) {
		this->ChangeUnitTypeAiActiveCount(type, 1);
		this->AiActiveUnitsByType[type].push_back(unit);
	}

	if (type->BoolFlag[TOWNHALL_INDEX].value) {
		this->NumTownHalls++;
	}
	
	for (const auto &[resource, quantity] : type->Stats[this->Index].get_resource_demands()) {
		this->change_resource_demand(resource, quantity);
	}
	
	if (this->AiEnabled && type->BoolFlag[COWARD_INDEX].value && !type->BoolFlag[HARVESTER_INDEX].value && !type->CanTransport() && type->Spells.size() == 0 && CMap::Map.Info.IsPointOnMap(unit->tilePos, unit->MapLayer) && unit->CanMove() && unit->Active && unit->GroupId != 0 && unit->Variable[SIGHTRANGE_INDEX].Value > 0) { //assign coward, non-worker, non-transporter, non-spellcaster units to be scouts
		this->Ai->Scouts.push_back(unit);
	}
	
	if (!type_change) {
		if (unit->get_character() != nullptr) {
			this->Heroes.push_back(unit);
		}

		if (!SaveGameLoading) {
			this->last_created_unit = unit;
		}
	}
}

void CPlayer::DecreaseCountsForUnit(CUnit *unit, const bool type_change)
{
	const wyrmgus::unit_type *type = unit->Type;

	this->ChangeUnitTypeCount(type, -1);
	
	wyrmgus::vector::remove(this->units_by_type[type], unit);

	if (this->units_by_type[type].empty()) {
		this->units_by_type.erase(type);
	}

	if (type->get_unit_class() != nullptr) {
		wyrmgus::vector::remove(this->units_by_class[type->get_unit_class()], unit);

		if (this->units_by_class[type->get_unit_class()].empty()) {
			this->units_by_class.erase(type->get_unit_class());
		}
	}
	
	if (unit->Active) {
		this->ChangeUnitTypeAiActiveCount(type, -1);
		
		wyrmgus::vector::remove(this->AiActiveUnitsByType[type], unit);
		
		if (this->AiActiveUnitsByType[type].empty()) {
			this->AiActiveUnitsByType.erase(type);
		}
	}
	
	if (type->BoolFlag[TOWNHALL_INDEX].value) {
		this->NumTownHalls--;
	}
	
	for (const auto &[resource, quantity] : type->Stats[this->Index].get_resource_demands()) {
		this->change_resource_demand(resource, -quantity);
	}
	
	if (this->AiEnabled && this->Ai && std::find(this->Ai->Scouts.begin(), this->Ai->Scouts.end(), unit) != this->Ai->Scouts.end()) {
		this->Ai->Scouts.erase(std::remove(this->Ai->Scouts.begin(), this->Ai->Scouts.end(), unit), this->Ai->Scouts.end());
	}
	
	if (!type_change) {
		if (unit->get_character() != nullptr) {
			this->Heroes.erase(std::remove(this->Heroes.begin(), this->Heroes.end(), unit), this->Heroes.end());
		}

		if (unit == this->last_created_unit) {
			this->last_created_unit = nullptr;
		}
	}
}

/**
**  Have unit of type.
**
**  @param type    Type of unit.
**
**  @return        How many exists, false otherwise.
*/
bool CPlayer::has_unit_type(const wyrmgus::unit_type *unit_type) const
{
	return this->GetUnitTypeCount(unit_type) > 0;
}

int CPlayer::get_population() const
{
	int people_count = 0;

	for (const auto &kv_pair : this->UnitTypesCount) {
		const wyrmgus::unit_type *unit_type = kv_pair.first;
		if (!unit_type->BoolFlag[ORGANIC_INDEX].value || unit_type->BoolFlag[FAUNA_INDEX].value) {
			continue;
		}

		people_count += kv_pair.second;
	}

	return static_cast<int>(pow(people_count, 2)) * wyrmgus::base_population_per_unit;
}

/**
**  Initialize the Ai for all players.
*/
void PlayersInitAi()
{
	for (int player = 0; player < NumPlayers; ++player) {
		if (CPlayer::Players[player]->AiEnabled) {
			AiInit(*CPlayer::Players[player]);
		}
	}
}

/**
**  Handle AI of all players each game cycle.
*/
void PlayersEachCycle()
{
	for (int player = 0; player < NumPlayers; ++player) {
		CPlayer *p = CPlayer::Players[player];
		
		if (p->LostTownHallTimer && !p->is_revealed() && p->LostTownHallTimer < ((int) GameCycle) && CPlayer::GetThisPlayer()->HasContactWith(*p)) {
			p->set_revealed(true);
			for (int j = 0; j < NumPlayers; ++j) {
				if (player != j && CPlayer::Players[j]->Type != PlayerNobody) {
					CPlayer::Players[j]->Notify(_("%s's units have been revealed!"), p->Name.c_str());
				} else {
					CPlayer::Players[j]->Notify("%s", _("Your units have been revealed!"));
				}
			}
		}
		
		
		for (size_t i = 0; i < p->Modifiers.size(); ++i) { //if already has the modifier, make it have the greater duration of the new or old one
			if ((unsigned long) p->Modifiers[i].second < GameCycle) {
				p->RemoveModifier(p->Modifiers[i].first); //only remove one modifier per cycle, to prevent too many upgrade changes from happening at the same cycle (for performance reasons)
				break;
			}
		}
		
		if (p->HeroCooldownTimer) {
			p->HeroCooldownTimer--;
		}

		if (p->AiEnabled) {
			AiEachCycle(*p);
		}
	}
}

/**
**  Handle AI of a player each second.
**
**  @param playerIdx  the player to update AI
*/
void PlayersEachSecond(int playerIdx)
{
	CPlayer *player = CPlayer::Players[playerIdx];

	if ((GameCycle / CYCLES_PER_SECOND) % 10 == 0) {
		for (const resource *resource : resource::get_all()) {
			int revenue = player->get_resource(resource) + player->get_stored_resource(resource) - player->get_last_resource(resource);
			//estimate per minute
			revenue *= 6;
			player->set_revenue(resource, revenue);
			player->set_last_resource(resource, player->get_resource(resource) + player->get_stored_resource(resource));
		}
	}
	if (player->AiEnabled) {
		AiEachSecond(*player);
	}

	player->UpdateFreeWorkers();
	//Wyrmgus start
	player->PerformResourceTrade();
	//Wyrmgus end
	player->update_current_quests();
}

/**
**  Handle AI of a player each half minute.
**
**  @param playerIdx  the player to update AI
*/
void PlayersEachHalfMinute(int playerIdx)
{
	CPlayer *player = CPlayer::Players[playerIdx];

	if (player->AiEnabled) {
		AiEachHalfMinute(*player);
	}
}

/**
**  Handle AI of a player each minute.
**
**  @param playerIdx  the player to update AI
*/
void PlayersEachMinute(int playerIdx)
{
	CPlayer *player = CPlayer::Players[playerIdx];

	if (player->AiEnabled) {
		AiEachMinute(*player);
	}

	player->update_quest_pool();
}

/**
**  Setup the player colors for the current palette.
**
**  @todo  FIXME: could be called before PixelsXX is setup.
*/
void SetPlayersPalette()
{
	for (int i = 0; i < PlayerMax - 1; ++i) {
		if (CPlayer::Players[i]->Faction == -1) {
			CPlayer::Players[i]->player_color = wyrmgus::player_color::get_all()[SyncRand(wyrmgus::player_color::get_all().size())];
		}
	}

	CPlayer::Players[PlayerNumNeutral]->player_color = wyrmgus::defines::get()->get_neutral_player_color();
}

/**
**  Notify player about a problem.
**
**  @param type    Problem type
**  @param pos     Map tile position
**  @param fmt     Message format
**  @param ...     Message varargs
**
**  @todo FIXME: We must also notfiy allied players.
*/
void CPlayer::Notify(int type, const Vec2i &pos, int z, const char *fmt, ...) const
{
	Assert(CMap::Map.Info.IsPointOnMap(pos, z));
	std::array<char, 128> temp{};
	uint32_t color;
	va_list va;

	// Notify me, and my TEAM members
	if (this != CPlayer::GetThisPlayer() && !IsTeamed(*CPlayer::GetThisPlayer())) {
		return;
	}

	va_start(va, fmt);
	temp[temp.size() - 1] = '\0';
	vsnprintf(temp.data(), temp.size() - 1, fmt, va);
	va_end(va);
	switch (type) {
		case NotifyRed:
			color = ColorRed;
			break;
		case NotifyYellow:
			color = ColorYellow;
			break;
		case NotifyGreen:
			color = ColorGreen;
			break;
		default: color = ColorWhite;
	}

	UI.get_minimap()->AddEvent(pos, z, color);

	if (this == CPlayer::GetThisPlayer()) {
		//Wyrmgus start
//		SetMessageEvent(pos, "%s", temp.data());
		SetMessageEvent(pos, z, "%s", temp.data());
		//Wyrmgus end
	} else {
		//Wyrmgus start
//		SetMessageEvent(pos, "(%s): %s", Name.c_str(), temp.data());
		SetMessageEvent(pos, z, "(%s): %s", Name.c_str(), temp.data());
		//Wyrmgus end
	}
}

/**
**  Notify player about a problem.
**
**  @param type    Problem type
**  @param pos     Map tile position
**  @param fmt     Message format
**  @param ...     Message varargs
**
**  @todo FIXME: We must also notfiy allied players.
*/
void CPlayer::Notify(const char *fmt, ...) const
{
	// Notify me, and my TEAM members
	if (this != CPlayer::GetThisPlayer() && !IsTeamed(*CPlayer::GetThisPlayer())) {
		return;
	}
	std::array<char, 128> temp{};
	va_list va;

	va_start(va, fmt);
	temp[temp.size() - 1] = '\0';
	vsnprintf(temp.data(), temp.size() - 1, fmt, va);
	va_end(va);
	if (this == CPlayer::GetThisPlayer()) {
		SetMessage("%s", temp.data());
	} else {
		SetMessage("(%s): %s", Name.c_str(), temp.data());
	}
}

void CPlayer::SetDiplomacyNeutralWith(const CPlayer &player)
{
	this->enemies.erase(player.Index);
	this->allies.erase(player.Index);

	//Wyrmgus start
	if (GameCycle > 0 && player.Index == CPlayer::GetThisPlayer()->Index) {
		CPlayer::GetThisPlayer()->Notify(_("%s changed their diplomatic stance with us to Neutral"), _(this->Name.c_str()));
	}
	//Wyrmgus end
}

void CPlayer::SetDiplomacyAlliedWith(const CPlayer &player)
{
	this->enemies.erase(player.Index);
	this->allies.insert(player.Index);
	
	if (GameCycle > 0 && player.Index == CPlayer::GetThisPlayer()->Index) {
		CPlayer::GetThisPlayer()->Notify(_("%s changed their diplomatic stance with us to Ally"), _(this->Name.c_str()));
	}
}

void CPlayer::SetDiplomacyEnemyWith(CPlayer &player)
{
	this->enemies.insert(player.Index);
	this->allies.erase(player.Index);
	
	if (GameCycle > 0) {
		if (player.Index == CPlayer::GetThisPlayer()->Index) {
			CPlayer::GetThisPlayer()->Notify(_("%s changed their diplomatic stance with us to Enemy"), _(this->Name.c_str()));
		} else if (this->Index == CPlayer::GetThisPlayer()->Index) {
			CPlayer::GetThisPlayer()->Notify(_("We have changed our diplomatic stance with %s to Enemy"), _(player.Name.c_str()));
		}
	}

	if (this->has_shared_vision_with(player)) {
		CommandSharedVision(this->Index, false, player.Index);
	}

	// if either player is the overlord of another (indirect or otherwise), break the vassalage bond after the declaration of war
	if (this->is_overlord_of(&player)) {
		player.set_overlord(nullptr, wyrmgus::vassalage_type::none);
	} else if (player.is_overlord_of(this)) {
		this->set_overlord(nullptr, wyrmgus::vassalage_type::none);
	}

	//if the other player has an overlord, then we must also go to war with them
	if (player.get_overlord() != nullptr) {
		this->SetDiplomacyEnemyWith(*player.get_overlord());
	}
}

void CPlayer::SetDiplomacyCrazyWith(const CPlayer &player)
{
	this->enemies.insert(player.Index);
	this->allies.insert(player.Index);
	
	if (GameCycle > 0 && player.Index == CPlayer::GetThisPlayer()->Index) {
		CPlayer::GetThisPlayer()->Notify(_("%s changed their diplomatic stance with us to Crazy"), _(this->Name.c_str()));
	}
}

void CPlayer::ShareVisionWith(const CPlayer &player)
{
	this->shared_vision.insert(player.Index);
	
	if (GameCycle > 0 && player.Index == CPlayer::GetThisPlayer()->Index) {
		CPlayer::GetThisPlayer()->Notify(_("%s is now sharing vision with us"), _(this->Name.c_str()));
	}
}

void CPlayer::UnshareVisionWith(const CPlayer &player)
{
	this->shared_vision.erase(player.Index);
	
	if (GameCycle > 0 && player.Index == CPlayer::GetThisPlayer()->Index) {
		CPlayer::GetThisPlayer()->Notify(_("%s is no longer sharing vision with us"), _(this->Name.c_str()));
	}
}

void CPlayer::set_overlord(CPlayer *overlord, const wyrmgus::vassalage_type)
{
	if (overlord == this->get_overlord()) {
		return;
	}

	if (overlord != nullptr && overlord->get_overlord() == this) {
		throw std::runtime_error("Cannot set player \"" + overlord->Name + "\" as the overlord of \"" + this->Name + "\", as the former is a vassal of the latter, and a vassal can't be the overlord of its own overlord.");
	}

	CPlayer *old_overlord = this->get_overlord();
	if (old_overlord != nullptr) {
		wyrmgus::vector::remove(old_overlord->vassals, this);

		//remove alliance and shared vision with the old overlord, and any upper overlords
		if (!SaveGameLoading) {
			while (old_overlord != nullptr) {
				this->break_overlordship_alliance(old_overlord);
				old_overlord = old_overlord->get_overlord();
			}
		}
	}

	this->overlord = overlord;
	this->vassalage_type = vassalage_type;

	if (overlord != nullptr) {
		overlord->vassals.push_back(this);

		//establish alliance and shared vision with the new overlord, and any upper overlords
		if (!SaveGameLoading) {
			while (overlord != nullptr) {
				this->establish_overlordship_alliance(overlord);
				overlord = overlord->get_overlord();
			}
		}
	}

	this->update_minimap_territory();
}

void CPlayer::establish_overlordship_alliance(CPlayer *overlord)
{
	this->SetDiplomacyAlliedWith(*overlord);
	overlord->SetDiplomacyAlliedWith(*this);
	CommandSharedVision(this->Index, true, overlord->Index);
	CommandSharedVision(overlord->Index, true, this->Index);

	//vassals should also be allied with overlords higher up in the hierarchy
	for (CPlayer *vassal : this->get_vassals()) {
		vassal->establish_overlordship_alliance(overlord);
	}
}

void CPlayer::break_overlordship_alliance(CPlayer *overlord)
{
	if (this->IsAllied(*overlord)) {
		this->SetDiplomacyNeutralWith(*overlord);
		overlord->SetDiplomacyNeutralWith(*this);
	}
	CommandSharedVision(this->Index, false, overlord->Index);
	CommandSharedVision(overlord->Index, false, this->Index);

	//vassals should also have their alliances with overlords higher up in the hierarchy broken
	for (CPlayer *vassal : this->get_vassals()) {
		vassal->break_overlordship_alliance(overlord);
	}
}

/**
**  Check if the player is an enemy
*/
bool CPlayer::IsEnemy(const CPlayer &player) const
{
	if (this->get_overlord() != nullptr && this->get_overlord()->IsEnemy(player)) {
		return true;
	}

	//be hostile to the other player if they are hostile, even if the diplomatic stance hasn't been changed
	return this->IsEnemy(player.Index) || player.IsEnemy(this->Index);
}

/**
**  Check if the unit is an enemy
*/
bool CPlayer::IsEnemy(const CUnit &unit) const
{
	if (
		unit.Player->Type == PlayerNeutral
		&& (unit.Type->BoolFlag[NEUTRAL_HOSTILE_INDEX].value || unit.Type->BoolFlag[PREDATOR_INDEX].value)
		&& this->Type != PlayerNeutral
	) {
		return true;
	}
	
	if (
		this != unit.Player
		&& this->Type != PlayerNeutral
		&& unit.CurrentAction() == UnitAction::Attack
		&& unit.CurrentOrder()->has_goal()
		&& unit.CurrentOrder()->get_goal()->Player == this
		&& !unit.CurrentOrder()->get_goal()->Type->BoolFlag[HIDDENOWNERSHIP_INDEX].value
	) {
		return true;
	}
	
	if (unit.Player->Index != this->Index && this->Type != PlayerNeutral && unit.Type->BoolFlag[HIDDENOWNERSHIP_INDEX].value && unit.IsAgressive() && !this->has_neutral_faction_type()) {
		return true;
	}
	
	return IsEnemy(*unit.Player);
}

/**
**  Check if the player is an ally
*/
bool CPlayer::IsAllied(const CPlayer &player) const
{
	//only consider yourself to be the ally of another player if they have the allied stance with you as well
	return this->IsAllied(player.Index) && player.IsAllied(this->Index);
}

/**
**  Check if the unit is an ally
*/
bool CPlayer::IsAllied(const CUnit &unit) const
{
	return IsAllied(*unit.Player);
}

bool CPlayer::IsVisionSharing() const
{
	return !this->shared_vision.empty();
}

bool CPlayer::has_shared_vision_with(const CPlayer &player) const
{
	return this->has_shared_vision_with(player.Index);
}

bool CPlayer::has_shared_vision_with(const CUnit &unit) const
{
	return this->has_shared_vision_with(*unit.Player);
}

bool CPlayer::has_mutual_shared_vision_with(const CPlayer &player) const
{
	return this->shared_vision.contains(player.Index) && player.shared_vision.contains(this->Index);
}

bool CPlayer::has_mutual_shared_vision_with(const CUnit &unit) const
{
	return this->has_mutual_shared_vision_with(*unit.Player);
}

/**
**  Check if the player is teamed
*/
bool CPlayer::IsTeamed(const CPlayer &player) const
{
	return Team == player.Team;
}

/**
**  Check if the unit is teamed
*/
bool CPlayer::IsTeamed(const CUnit &unit) const
{
	return IsTeamed(*unit.Player);
}

//Wyrmgus start
/**
**  Check if the player has contact with another (used for determining which players show up in the player list and etc.)
*/
bool CPlayer::HasContactWith(const CPlayer &player) const
{
	return player.StartMapLayer == this->StartMapLayer || (player.StartMapLayer < (int) CMap::Map.MapLayers.size() && this->StartMapLayer < (int) CMap::Map.MapLayers.size() && CMap::Map.MapLayers[player.StartMapLayer]->world == CMap::Map.MapLayers[this->StartMapLayer]->world && CMap::Map.MapLayers[player.StartMapLayer]->plane == CMap::Map.MapLayers[this->StartMapLayer]->plane);
}

/**
**  Check if the player's faction type is a neutral one
*/
bool CPlayer::has_neutral_faction_type() const
{
	if (
		this->get_civilization() != nullptr
		&& this->get_faction() != nullptr
		&& (is_faction_type_neutral(this->get_faction()->get_type()))
	) {
		return true;
	}

	return false;
}

/**
**  Check if the player can use the buildings of another, for neutral building functions (i.e. unit training)
*/
bool CPlayer::has_building_access(const CPlayer *player, const ButtonCmd button_action) const
{
	if (player->IsEnemy(*this)) {
		return false;
	}

	if (player->Type == PlayerNeutral) {
		return true;
	}

	if (player->get_faction() == nullptr) {
		return false;
	}

	if (
		player->has_neutral_faction_type()
		&& (player->get_overlord() == nullptr || this->is_any_overlord_of(player) || player->get_overlord()->IsAllied(*this))
	) {
		if (player->get_faction()->get_type() != wyrmgus::faction_type::holy_order || (button_action != ButtonCmd::Train && button_action != ButtonCmd::TrainClass && button_action != ButtonCmd::Buy) || wyrmgus::vector::contains(this->Deities, player->get_faction()->get_holy_order_deity())) { //if the faction is a holy order, the player must have chosen its respective deity
			return true;
		}
	}

	return false;
}

bool CPlayer::has_building_access(const CPlayer *player) const
{
	return this->has_building_access(player, ButtonCmd::None);
}

bool CPlayer::has_building_access(const CUnit *unit, const ButtonCmd button_action) const
{
	if (unit->IsEnemy(*this)) {
		return false;
	}

	return this->has_building_access(unit->Player, button_action);
}

bool CPlayer::has_building_access(const CUnit *unit) const
{
	return this->has_building_access(unit, ButtonCmd::None);
}

bool CPlayer::HasHero(const wyrmgus::character *hero) const
{
	if (!hero) {
		return false;
	}
	
	for (const CUnit *hero_unit : this->Heroes) {
		if (hero_unit->get_character() == hero) {
			return true;
		}
	}
	
	return false;
}

void NetworkSetFaction(int player, const std::string &faction_name)
{
	const wyrmgus::faction *faction = wyrmgus::faction::try_get(faction_name);
	int faction_id = faction ? faction->ID : -1;
	SendCommandSetFaction(player, faction_id);
}

bool IsNameValidForWord(const std::string &word_name)
{
	if (word_name.empty()) {
		return false;
	}
	
	if (
		word_name.find('\n') != -1
		|| word_name.find('\\') != -1
		|| word_name.find('/') != -1
		|| word_name.find('.') != -1
		|| word_name.find('*') != -1
		|| word_name.find('[') != -1
		|| word_name.find(']') != -1
		|| word_name.find(':') != -1
		|| word_name.find(';') != -1
		|| word_name.find('=') != -1
		|| word_name.find(',') != -1
		|| word_name.find('<') != -1
		|| word_name.find('>') != -1
		|| word_name.find('?') != -1
		|| word_name.find('|') != -1
	) {
		return false;
	}
	
	if (word_name.find_first_not_of(' ') == std::string::npos) {
		return false; //name contains only spaces
	}
	
	return true;
}
//Wyrmgus end
